/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: audio.cpp,v $
   Revision 1.5  2007/12/15 05:02:14  wb6ymh
   Added code to reset pGSM after compression loop. Where did it go?  Broke
   EchoLink and RTP/GSM transmission rather badly!

   Revision 1.3  2007/12/07 23:42:56  wb6ymh
   1. Added support for 8 bit PCM .wav files.
   2. Moved ADPCM globals into CPort class.
   3. Added exit after call to execvp in StartNextAnnoucement so we don't end
      up with two copies of tlb attempting to run if TTS_ExePath doesn't point
      to a runnable executable.
   4. Replaced NetQueueLen and MaxNetQueueLen with new configuration variable
      VoipDejitterDelay.  Number of packets of buffering is now calculated
      based on the amount of audio contained in the packets.  There's a
      lot more audio in an SF/GSM packet than in a RTP/GSM packet.
   5. Completely changed they way the VoIP courtesy tone is generated. It's
      now based on calculations of when the audio that's been sent to the
      sound device runs out.  Flaws in the old approarch became apparent when
      the old excessively long Internet buffer was reduced.
   6. Changed the way the Tx timeout was handled.  Previously VoIP queues were
      not read during a timeout which made the first VoIP timeout final since
      an empty queue was a condition for resetting the timeout.
   7. Moved transmitter keying from Audio_Handler to WriteAudioOut to ensure
      transmitter is always keyed when audio is written.  Previously certain
      timing cases (like the first .tonegen command after startup) would not
      key the transmitter.
   8. Corrected possible sign extension bug in ADPCM code when packing state
      into stream.
   9. Corrected buffer overwrite bug in ADPCM transmissions that caused severe
      clicks and audio distortion.
   10. Reduced the size of the audio driver's buffer from 15 x 256 to 5 x 256
       to limit maximum delays.
   11. Modified CmdId so it works when IdInterval is 0.
   12. Fixed infinite loop in AddToneGen that occurred when a tonegenerator was
       added while a different one was running.
   13. Added CtoneResetsTimeout to allow transmit timeout timer to be reset by
       courtesy tone rather than waiting for transmitter unkey.
   14. Corrected bug in VoipAudioAvailable while allowed audio to be consumed
       during the buffering state.  Caused audio dropouts at beginning of
       ADPCM transmissions.
   15. Removed logging of failed execvp calls from CmdScript ... it's after
       a fork and the logging routines are probably not thread safe.

   Revision 1.2  2007/12/01 06:39:15  wb6ymh
   1. Removed STRCMP macro, its in common.h now.
   2. Removed sweep tone generator code and replaced it with calls to the
      general purpose CToneGen class.
   3. Changed almost all DPRINTFs to D2PRINTFs to quiet the output in
      DebugLevel 1.
   4. Modified CPort::KeyTx to check SerialFD before using it. Gets rid of
      some bogus error messages when SerialInit fails.
   5. Added support for  courtesy tones, transmitter hang time, and a
      transmit timeout timer.
   6. Deleted dead code.WriteAudioOut, and CPort ActiveSources.
   7. Added support for general purpose tone generators.
   8. Replaced fflush(stdout) alls with DFLUSH() macros (which checks debug
      mode before calling fflush(stdio). [Trival optimization]
   9. Corrected error logging for failure to open serial device in SerialInit.
   10. Added code to set CWChar to NULL when CW message completes.  Fixes
       stuck transmitter following CW ID.
   11. Added CmdToneGen, CmdDtmfGen.
   12. Removed left over debug code that linked port 1 to port 0 on
       initialization.  Fixes crashes when link command was run on a single
       port system.

   Revision 1.1  2007/11/26 07:24:30  wb6ymh
   1. Added parallel port for FreeBSD.
   2. Massive modifications to convert most routines into C++.  Makes supporting
      multiple ports easer since the code wasn't orginally written with that
      in mind.
   3. Modified FillAudioBuf to source audio only from VoIP sources that have
      this port's bit set in the new TxMap variable.
   4. Added new code to FillAudioBuf to source audio from ports which have
      the corresponding bit set in the new RxMap.
   5. Modified AdpcmTx and GsmTx  to only send data to VoIP clients that have
      the port's bit set in the new RxMap variable.
   6. Added support for RxCosMethod 7 to PollCOS.
   7. Modified PollDTMF to use new DtmfMethod rather than TxKeyMethod to
      select type of DTMF handling.
   8. Added support for PTT_bit and COS_Bit configuration variables.
   9. Added .link and .unlink commands.

   Revision 1.2  2007/11/21 16:27:34  wb6ymh
   1. Added code to share IRLP interface with other code. (Enabled by the
      ShareIRLP_Port configuration file variable.
   2. Added support for TxMethod 5 (PTT via iMic GPIO).
   3. Added support for RxMethod 6 (COS via Linux event system)

   Revision 1.1  2007/06/30 00:04:25  wb6ymh
   Initial import - source files for thelinkbox.

   
*/ 
#include "common.h"

// FreeBSD, Linux, etc..
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
   #include <sys/time.h>
   #include <time.h>
#else
   #ifdef HAVE_SYS_TIME_H
      #include <sys/time.h>
   #else
      #include <time.h>
   #endif
#endif
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include <sys/soundcard.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
#include <netdb.h>
#include <ctype.h>
#ifdef HAVE_SYS_TIMEB_H
   #include <sys/timeb.h>
#endif
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <math.h>

#ifdef HAVE_SYS_SYSLIMITS_H
#include <sys/syslimits.h> // for PATH_MAX
#endif

#include <sys/file.h>


#include "rtp.h"
#include "avl.h"
#include "main.h"
#include "configvars.h"
#include "configfile.h"
#include "dirclient.h"
#include "conference.h"
#include "audio.h"
#include "users.h"
#include "eventhook.h"
#include "ctcss.h"
#include "tbd.h"
#include "devinput.h"

#ifdef USB_SUPPORT
#include "usbdev.h"
#endif
#include "ports.h"


// IRLP interface board equates
#define IRLP_LOCKFILE   "/tmp/irlp-lockfile-parport0"

#define TTS_OUT_FILENAME   "tts.wav"
#define TTS_IN_FILENAME    "tts.in"

#define IRLP_PTT        0x02
#define IRLP_AUX1       0x04
#define IRLP_AUX2       0x08

#define IRLP_DTMF_SHIFT 3
#define IRLP_COS        0x80  // Busy

#ifdef __linux__
#include <linux/ppdev.h>
#include <linux/parport.h>
#define PP_RD_DATA   PPRDATA
#define PP_WR_DATA   PPWDATA
#define PP_RD_STATUS PPRSTATUS
#define PP_DATA_TYPE char
#define IRLP_DEVICE     "/dev/parport0"
#endif

#ifdef __FreeBSD__
#include <dev/ppbus/ppi.h>
#include <dev/ppbus/ppbconf.h>
#define PP_RD_DATA   PPIGDATA
#define PP_WR_DATA   PPISDATA
#define PP_RD_STATUS PPIGSTATUS
#define PP_DATA_TYPE int
#define IRLP_DEVICE     "/dev/ppi0"
#endif

/*
bytes      variable      description
0  - 3     'RIFF'/'RIFX' Little/Big-endian
4  - 7     wRiffLength   length of file minus the 8 byte riff header
8  - 11    'WAVE'
12 - 15    'fmt '
16 - 19    wFmtSize       length of format chunk minus 8 byte header 
20 - 21    wFormatTag     identifies PCM, ULAW etc
22 - 23    wChannels      
24 - 27    dwSamplesPerSecond  samples per second per channel
28 - 31    dwAvgBytesPerSec    non-trivial for compressed formats
32 - 33    wBlockAlign         basic block size
34 - 35    wBitsPerSample      non-trivial for compressed formats
*/

typedef struct {
   char  Riff[4];             // 'RIFF'/'RIFX' Little/Big-endian
   int   wRiffLength;         // length of file minus the 8 byte riff header
   char  Wave[4];             // 'WAVE'
   char  Fmt[4];              // 'fmt '
   int   wFmtSize;            // length of format chunk minus 8 byte header 
   short wFormatTag;          // identifies PCM, ULAW etc
   short wChannels;
   int   dwSamplesPerSecond;  // samples per second per channel
   int   dwAvgBytesPerSec;    // non-trivial for compressed formats
   short wBlockAlign;         // basic block size
   short wBitsPerSample;      // non-trivial for compressed formats
} WaveHdr;

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif  /* min */

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif  /* min */

#define CORE_DUMP()  \
   LOG_ERROR(("%s#%d: dumping core\n",__FUNCTION__,__LINE__)); \
   DumpMe()


#ifndef  M_PI
#define  M_PI     3.14159265358979323846  /* pi */
#endif

#define SERIAL_OPEN_FLAGS O_FSYNC | O_APPEND | O_NONBLOCK | O_RDWR | \
        O_NOCTTY | O_NDELAY

#ifndef  DEBUG_PCM_IN
#define AUDIO_OPEN_FLAGS   O_RDWR | O_NONBLOCK
#else
#define AUDIO_OPEN_FLAGS   O_RDWR
#endif

// global bit map of ports with active signals
int CosMap = 0;


#define MAX_MIXED_LEVEL 32000


int MixBuffer[AUDIO_BUF_SIZE];

int16 ToneBuf[AUDIO_BUF_SIZE];

int16 PlayBuf[AUDIO_BUF_SIZE];


int WriteWaits = 0;
int WriteWakeups = 0;
int WriteStalls = 0;

int IRLPfd = -1;
int LockIRLPfd = -1;
int gIRLP_PortLocked = FALSE;
int bTTS_Running = FALSE;
int g_bRxCTCSS = FALSE;

FILE *fp = NULL;

int DitSamples;
int CwRampCount;     // number of samples ramping up and down from full level
int ProblemSeq = 10805;



struct BaudRateTbl_TAG {
   int   Baudrate;
   int   SpeedConstant;
} BaudRateTbl[] = {
   {110,B110},
   {300,B300},
   {600,B600},
   {1200,B1200},
   {2400,B2400},
   {4800,B4800},
   {9600,B9600},
   {19200,B19200},
   {0}
};

CPort *Ports[MAX_PORTS];

void AnnounceConnect(char *Args);
void AnnounceDisconnect(char *Args);
void SendCWMsg(char *Msg);
void ID();
void PlayAnnouncementFile(char *FileName,int Flags);
void ExecuteCommand(char *Command);

struct EventHandler {
   char *EventDesc;
   void (*EventHandler)(char *Args);
} EventHandlers[] = {
   { "connected", AnnounceConnect},
   { "disconnected", AnnounceDisconnect},
   { NULL } // end of table
};



/* A -> Z */
char *CwAlphaLookup[] = {
   ".-","-...","-.-.","-..",".","..-.","--.","....","..",".---","-.-",
   ".-..","--","-.","---",".--.","--.-",".-.","...","-","..-","...-",
   ".--","-..-","-.--","--.."
};

/* 0 -> 9 */
char *CwNumericLookup[] = {
   "-----",".----","..---","...--","....-",".....","-....",
   "--...","---..","----."
};

struct CwMiscLookupTAG {
   char  Char;
   char  *Morse;
} CwMiscLookup[] = {
   {'?', "..--.."},
   {'/', "-..-."},
   {'.', ".-.-.-"},
   {',', "--..--"},
   {'-', "-...-"},   // aka BT
   {0}   // end of table
};

void Breakpoint()
{
   printf("Breakpoint called.\n");
}
/*
   return if a <= b < c in a circular sense.
   (From Tanenbaum 1981, pg 158)
*/
int Between(u_int16 a, u_int16 b, u_int16 c)
{
   if( (a <= b && b < c) || (c < a && a <= b) || (b < c && c < a) ) {
      return TRUE;
   }
   return FALSE;
}


void EndPointEvent(char *EventText)
{
   int Len;
   char *cp;
   int i;

   for(i = 0; EventHandlers[i].EventDesc != NULL; i++) {
      Len = strlen(EventHandlers[i].EventDesc);
      if(strncmp(EventText,EventHandlers[i].EventDesc,Len) == 0) {
         if((cp = strchr(EventText,' ')) != NULL) {
            cp++;
         }

         EventHandlers[i].EventHandler(cp);
         break;
      }
   }
}

int ClaimIRLP() 
{
   int Ret = 0;   // Assume the best
   if(ShareIRLP_Port) {
      if(flock(LockIRLPfd,LOCK_EX) != 0) {
         Ret = errno;
         LOG_ERROR(("%s: flock(LOCK_EX) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      }
#ifdef __linux__
      else if(ioctl(IRLPfd,PPCLAIM) != 0) {
         Ret = errno;
         LOG_ERROR(("%s: ioctl(PPCLAIM) failed: %s",__FUNCTION__,
                    Err2String(errno)));
         if(flock(LockIRLPfd,LOCK_UN) != 0) {
            LOG_ERROR(("%s: flock(LOCK_UN) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
      }
#endif
      else {
         gIRLP_PortLocked = TRUE;
      }
   }

   return Ret;
}

void ReleaseIRLP()
{
   if(ShareIRLP_Port && gIRLP_PortLocked) {
#ifdef __linux__
      if(ioctl(IRLPfd,PPRELEASE) != 0) {
         LOG_ERROR(("%s: ioctl(PPRELEASE) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      } else 
#endif
      if(flock(LockIRLPfd,LOCK_UN) != 0) {
         LOG_ERROR(("%s: flock(LOCK_UN) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      }
      else {
         gIRLP_PortLocked = FALSE;
      }
   }
}

CPort::CPort()
{
// set default configuration variable values
   VoxThreshold = 0;
   VoxHoldTime = 0;
   VoxTripDelay = 0;
   TxKeyMethod = 0;
   RxCosMethod = 0;
   DtmfMethod = 0;
   SerialBaudRate = 2400;
   DTMFPollTime = 50;
   InvertPTT = 0;
   InvertCOS = 0;
   CWToneLevel = 5000;
   CWSpeed = 20;
   CWAutoJam = 0;
   IdInterval = 0;
   VoiceIDIdleTime = 0;
   HalfDuplexRF = 0;
   HalfDuplex = 0;
   PCMRate = 8000;
   CTCSSLevel = 0;
   Twist = -1;
   ReverseTwist = -1;
// set CtcssDecoderSamples to 300 milliseconds decode time @ 8 Khz sample rate.
   CtcssDecoderSamples = 3200;  
   DecodeCTCSS = 0;
   MuteDTMF = 1;
   bMonitor = 0;
   COS_KeyCode = 0;
   AudioIn8Bit = 0;
   CWToneFreq = 1000;

   EchoCTCSSFreq = 0.0;

   SerialDevName = NULL;
   CWId = NULL;
   VoiceID = NULL;
   AudioDevice = NULL;
   AudioDongleSN = NULL;
   PortName = NULL;

// Init class variables
   DTMF_Digits = 0;
   Playfp = NULL;
   pPlayHead = NULL;
   LastMixGain = 1024;  // 1:1
   CTCSSAngle = 0.0;
   CWChar = NULL;
   ElementCount = 0;
   bCwToneOn = FALSE;
   bElementSpace = FALSE;
   RampCount = 0;
   Element = NULL;
   AudioInAvailable = 0;
   AudioInBuf8 = NULL;
   AudioInBuf = NULL;
   bDevInCOS = FALSE;
   g_bCosActive = FALSE;
   pAudioC = NULL;
   RxMap = 0;
   RxAudioReady = 0;
   bDiscFeed = 0;
   bUngatedRx = 0;
   pDTMFCommands = NULL;
   bTossRxIn = 0;
   PTT_Bit = -1;
   COS_Bit = -1;
   pSerialSlave = NULL;
   pParallelSlave = NULL;
   SerialFd = -1;
   ToneGenerators = NULL;
   TimeoutToneGen = NULL;
   Ctone = 0;
   CtoneMap = 0;
   CtoneDelay = 0;
   CtoneTimer.tv_sec = 0;
   VoipCtoneTimer.tv_sec = 0;
   HangTimer.tv_sec = 0;
   TxTimeout = 0;
   TimeoutTone = 0;
   TxKeyTime = 0;
   bTimeOut = 0;
   CtoneResetsTimeout = 0;
   ADPCM_Bytes = 0;
   AudioAvailable = 0;
   AudioUsed = 0;
   hGSM_Encode = NULL;
   TxSequence = 0;
   RtpTxSequence = 0;
   GsmFramesAvail = 0;
#ifdef USB_SUPPORT
   udev = NULL;
   UsbDevType = 0;
#endif
};


void CPort::KeyTx(int bKey)
{
   int ModemCtrl;
   PP_DATA_TYPE IRLP_Bits;

   bTxKeyed = bKey;

   if(bKey) {
      D2PRINTF(("K"));
      TxKeyTime = TimeNow.tv_sec;
      if((TimeNow.tv_sec - LastID) > IdInterval) {
      // Time to ID
         ID();
      }
   }
   else {
      D2PRINTF(("U"));
   }

   switch(TxKeyMethod) {
      case 1:  // WB2REM / VA3TO interface
      // nothing to do now, tx keying is handed in PollDTMF()
         break;

      case 2:  // Parallel port interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_DATA,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_DATA) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else {
            if((bKey && !InvertPTT) || (!bKey && InvertPTT)) {
               IRLP_Bits |= PTT_Bit;
            }
            else {
               IRLP_Bits &= ~PTT_Bit;
            }

            if(ioctl(IRLPfd,PP_WR_DATA,&IRLP_Bits)) {
               LOG_ERROR(("%s: ioctl(PP_WR_DATA) failed: %s",__FUNCTION__,
                          Err2String(errno)));
            } 
         }
         ReleaseIRLP();
         break;

      case 3:  // RTS
      case 4:  // DTR
         if(SerialFd != -1) {
            if(ioctl(SerialFd,TIOCMGET,&ModemCtrl) == -1) {
               LOG_ERROR(("KeyTx(): ioctl(TIOCMGET) failed: %s",
                          Err2String(errno)));
            }
            else {
               if((bKey && !InvertPTT) || (!bKey && InvertPTT)) {
                  ModemCtrl |= PTT_Bit;
               }
               else {
                  ModemCtrl &= ~PTT_Bit;
               }
               if(ioctl(SerialFd,TIOCMSET,&ModemCtrl) == -1) {
                  LOG_ERROR(("KeyTx(): ioctl(TIOCMSET) failed: %s",
                             Err2String(errno)));
               }
            }
         }
         break;

      case 5:  // iMic GPIO
#ifdef USB_SUPPORT
         UsbKeyRadio(bKey);
#endif
         break;
   }
}

void TTSCB(int Pid, int Status)
{
// The TTS program has exited...
   bTTS_Running = FALSE;
}

void CPort::FreeAnnoucement()
{
   PlayPCM *p;

   if(Playfp != NULL) {
      fclose(Playfp);
      Playfp = NULL;
   }

   if((p = pPlayHead) != NULL) {
      pPlayHead = p->Link;
      free(p->FileName);
      free(p);
   }
}

void CPort::StartNextAnnoucement()
{
   PlayPCM *p = pPlayHead;
   FILE *TTSInFp = NULL;
   char Filename[PATH_MAX];
   int Bytes2Write;
   char *argv[4];
   char OutFilename[PATH_MAX];
   pid_t TTSPid = -1;
   
   while((p = pPlayHead) != NULL) {
      if(p->Flags & PLAY_FLAG_TTS) {
         if(TTSInFp == NULL) {
         // Open TTS input file
            snprintf(Filename,PATH_MAX,"%s/%s",WorkingDir,TTS_IN_FILENAME);
            if((TTSInFp = fopen(Filename,"w")) == NULL) {
               LOG_ERROR(("%s#%d: Couldn't open \"%s\", %s",__FUNCTION__,
                          __LINE__,Filename,Err2String(errno)));
            }
         }
         else {
         // File already open, append a space so we don't run words together
            if(fputs(" ",TTSInFp) == EOF) {
               LOG_ERROR(("%s: Error reading \"%s\", %s",__FUNCTION__,
                          Filename,Err2String(errno)));
               fclose(TTSInFp);
               TTSInFp = NULL;
            }
         }

         if(TTSInFp != NULL) {
            Bytes2Write = strlen(p->FileName);
            if(fwrite(p->FileName,Bytes2Write,1,TTSInFp) != 1) {
               LOG_ERROR(("%s: Error reading \"%s\", %s",__FUNCTION__,
                          Filename,Err2String(errno)));
               fclose(TTSInFp);
               TTSInFp = NULL;
            }
         }
         FreeAnnoucement();
      }
      else {
         if(TTSInFp != NULL) {
         // Don't start PCM file playback until TTS finishes
            break;
         }
         if((Playfp = fopen(p->FileName,"r")) == NULL) {
            LOG_ERROR(("%s#%d: Couldn't open \"%s\", %s",__FUNCTION__,__LINE__,
                       p->FileName,Err2String(errno)));
            FreeAnnoucement();
         }
         else {
            WaveHdr Hdr;
         // read 44 byte WAV header
            if(fread(&Hdr,sizeof(Hdr),1,Playfp) != 1) {
               LOG_ERROR(("%s: fread failed reading %s - %s\n",__FUNCTION__,
                          p->FileName,Err2String(errno)));
               FreeAnnoucement();
            }
            else {
            // Stupid test for now... fix me someday
               if(Hdr.wBitsPerSample == 8) {
                  b8BitFile = TRUE;
                  break;
               }
               else if(Hdr.wBitsPerSample == 16) {
                  b8BitFile = FALSE;
                  break;
               }
               else {
                  LOG_ERROR(("%s: %s has unsupported wBitsPerSample (%d)\n",
                             __FUNCTION__,p->FileName,Hdr.wBitsPerSample));
                  FreeAnnoucement();
               }
            }
         }
      }
   }

   if(TTSInFp != NULL) {
   // An TTS input file has been written, start the TTS program to 
   // render it to audio
      fclose(TTSInFp);

      snprintf(OutFilename,PATH_MAX,"%s/%s",WorkingDir,TTS_OUT_FILENAME);
      unlink(OutFilename);
   // Create a new announcement entry at the front of the queue for playback

      PlayAnnouncementFile(OutFilename,PLAY_FLAG_TTS_OUT);

   // NB: make sure all buffered streams are written to disk,
   // otherwise the data will be written twice, once by the main
   // process and once by the forked process !

      if(fflush(NULL)) {
         LOG_ERROR(("%s: fflush failed %s",__FUNCTION__,Err2String(errno)));
      }

      bTTS_Running = TRUE;
      if((TTSPid = fork()) < 0) {
         FreeAnnoucement();
         bTTS_Running = FALSE;
         LOG_ERROR(("%s: fork failed %s",__FUNCTION__,Err2String(errno)));
      }
      else if(TTSPid == 0) {
      // child
         argv[0] = TTS_ExePath;
         argv[1] = Filename;
         argv[2] = OutFilename;
         argv[3] = NULL;
         execvp(TTS_ExePath,argv);
      // NB: if execvp fails, BAIL!
         exit(1);
      }
      else {
      // Parent
         RegisterExitCB(TTSPid,TTSCB);
      }
   }

   if(pPlayHead != NULL && !(pPlayHead->Flags & PLAY_FLAG_ID)) {
   // This annoucement is not a voice ID so set the LastActivity
   // timer for the ID interval
      LastActivity = TimeNow.tv_sec;
   }
}


void CPort::PlayAnnouncementFile(char *FileName,int Flags)
{
   PlayPCM *p;

   if((p = (PlayPCM *) malloc(sizeof(PlayPCM))) == NULL) {
      LOG_ERROR(("PlayAnnouncementFile(): Error: out of memory\n"));
   }
   else {
      p->FileName = strdup(FileName);
      p->Flags = Flags;
      if(Flags & PLAY_FLAG_TTS_OUT) {
      // put on head of queue
         p->Link = pPlayHead;
         pPlayHead = p;
         if(p->Link == NULL) {
            pPlayTail = p;
         }
      }
      else {
      // put on tail of queue
         p->Link = NULL;
         if(pPlayHead == NULL) {
            pPlayHead = p;
         }
         else {
            pPlayTail->Link = p;
         }
         pPlayTail = p;
      }
      if(Playfp == NULL && !(Flags & PLAY_FLAG_TTS_OUT)) {
         StartNextAnnoucement();
      }
   }
}

// Say the specified phrase
void CPort::SayPhrase(char *Phrase)
{
   PlayPCM *p;
   char FileName[PATH_MAX];
   char *cp;
   char *Temp;
   struct stat FileStats;

   FileName[0] = 0;

   if((cp = strstr(Phrase,".wav")) != NULL && strlen(cp) == 4) {
   // a filename was specified
      strcpy(FileName,Phrase);
   }
   else if(TTS_ExePath == NULL) {
   // Lookup PCM file that contains the Phrase we want to say
      if(strlen(Phrase) == 1) {
      // Special case just a single character
         snprintf(FileName,sizeof(FileName),"pcm/%c.wav",toupper(*Phrase));
      }
      else {
      // replace all spaces with underscores
         Temp = strdup(Phrase);
         cp = Temp;
         while((cp = strchr(cp,' ')) != NULL) {
            *cp++ = '_';
         }
         snprintf(FileName,sizeof(FileName),"pcm/%s.wav",Temp);
         free(Temp);
         if(stat(FileName,&FileStats) != 0) {
         // no such file
            FileName[0] = 0;
            LOG_ERROR(("%s: Unable to find file for phrase \"%s\"\n",
                       __FUNCTION__,Phrase));
         }
      }
   }

   if(TTS_ExePath != NULL || FileName[0] != 0) {
      if((p = (PlayPCM *) malloc(sizeof(PlayPCM))) == NULL) {
         LOG_ERROR(("%s: Error: out of memory\n",__FUNCTION__));
      }
      else {
         p->Link = NULL;
         if(pPlayHead == NULL) {
            pPlayHead = p;
         }
         else {
            pPlayTail->Link = p;
         }
         pPlayTail = p;
         if(FileName[0] != 0) {
            p->Flags = 0;
            p->FileName = strdup(FileName);
            if(Playfp == NULL) {
               StartNextAnnoucement();
            }
         }
         else {
            p->Flags = PLAY_FLAG_TTS;
            p->FileName = strdup(Phrase);
         }
      }
   }
   if(!pAudioC->bWriteWait && Playfp != NULL) {
      WriteAudioOut(pAudioC);
   }
}

void CPort::SayCall(char *Call)
{
   int i;
   char FileName[80];
   int CallLen = strlen(Call);
   char Temp[80];
   char  *cp;
   struct stat FileStats;
   char AlphaNumeric[2];

// Try for a exact match filename

   snprintf(Temp,sizeof(Temp),"%s",*Call == '*' ? &Call[1] : Call);
   if((cp = strchr(Temp,'*')) != NULL) {
   // remove trailing '*'
      *cp = 0;
   }
   if((cp = strchr(Temp,' ')) != NULL) {
   // remove trailing spaces
      *cp = 0;
   }
   snprintf(FileName,sizeof(FileName),"pcm/%s.wav",Temp);
   if(stat(FileName,&FileStats) == 0) {
   // we have a file for this callsign/conference name, play it.
      PlayAnnouncementFile(FileName,0);
   }
   else {
   // Don't have an exact match, spell the call letter by letter
      for(i = 0; i < CallLen; i++) {
         if(isalnum(Call[i])) {
            AlphaNumeric[0] = Call[i];
            AlphaNumeric[1] = 0;
            SayPhrase(AlphaNumeric);
         }
         else if(Call[i] == '*') {
            if(i != 0) {
               SayPhrase("conference");
            }
         }
         else if(Call[i] == '-') {
            i++;
            if(toupper(Call[i]) == 'L') {
            // Link
               SayPhrase("link");
            }
            else if(toupper(Call[i]) == 'R') {
            // Repeater
               SayPhrase("repeater");
            }
         }
         else if(Call[i] == ' ') {
            break;
         }
      }
   }

   if(!pAudioC->bWriteWait) {
      WriteAudioOut(pAudioC);
   }
}

CPort *GetPortPtrF(int Port,int Line)
{
   CPort *p = NULL;
   if(Port < 0 || Port >= MAX_PORTS || (p = Ports[Port]) == NULL) {
      LOG_ERROR(("%s#%d: dumping core\n",__FUNCTION__,Line));
      DumpMe();
   }

   return p;
}

#define GetPortPtr(x) GetPortPtrF(x,__LINE__)

// "connected <protocol> <callsign> <number of connections>"
void AnnounceConnect(char *Args)
{
   char *cp;
   CPort *pPort = GetPortPtr(CurrentPort);

   if((cp = strchr(Args,' ')) != NULL) {
      pPort->SayPhrase("connected to");
      pPort->SayCall(cp+1);
   }
}

// disconnected <reason> <callsign> <number of connections left>"
void AnnounceDisconnect(char *Args)
{
   char *cp;
   char *cp1;
   CPort *pPort = GetPortPtr(CurrentPort);

   if((cp = strchr(Args,' ')) != NULL) {
      cp++;
      if((cp1 = strchr(cp,' ')) != NULL) {
         *cp1 = 0;
      }

      if(StationDisconnected != NULL) {
      // suppress the announcement if we explicitly disconnected this guy
         if(STRCMP(StationDisconnected,cp) == 0) {
         // We disconnected him
            cp = NULL;
            free(StationDisconnected);
            StationDisconnected = NULL;
         }
      }
   }

   if(cp != NULL) {
      pPort->SayCall(cp);
      pPort->SayPhrase("disconnected");
   }
}

void CPort::ID()
{
   int TimeSinceLastActvity = TimeNow.tv_sec - LastActivity;

   if(!bTxKeyed && HalfDuplexRF && 
      (bTalking || TimeLapse(&LastSpoke) < 5000)) 
   {  // We're running on a half duplex link and we're talking or have been
      // talking recently.  Don't key the transmitter !
   }
   else {
      if(CWAutoJam || TimeSinceLastActvity < IdInterval) {
      // Send an ID
         LastID = TimeNow.tv_sec;

         if(VoiceID != NULL && TimeSinceLastActvity > VoiceIDIdleTime) {
         // We have a voice ID and the channel has been idle long enough 
         // to play it
            PlayAnnouncementFile(VoiceID,PLAY_FLAG_ID);
         }
         else if(CWId != NULL) {
            SendCWMsg(CWId);
         }
      }
      else {
      // Don't ID again until we have some new activity
         LastID = 0;
      }
   }
}


int CPort::GetAnnounceAudio()
{
   int Ret = FALSE;
   int Read;
   int Bytes2Read;
   
   PlayAvail = 0;

   while(!bTTS_Running && pPlayHead != NULL && PlayAvail < AUDIO_BUF_SIZE) {
      if(Playfp == NULL) {
         StartNextAnnoucement();
      }

      if(Playfp != NULL) {
         if(b8BitFile) {
            unsigned char TempBuf[AUDIO_BUF_SIZE];
            int i;

            Bytes2Read = sizeof(TempBuf) - PlayAvail;
            Read = fread(TempBuf,1,Bytes2Read,Playfp);
            if(Read > 0) {
            // Convert to 16 bit samples in PlayBuf
               for(i = 0; i < Read; i++) {
                  PlayBuf[PlayAvail++] = (TempBuf[i] - 0x80) << 8;
               }
            }
         }
         else {
            Bytes2Read = sizeof(PlayBuf) - (PlayAvail * sizeof(int16));
            Read = fread(&PlayBuf[PlayAvail],1,Bytes2Read,Playfp);
            if(Read > 0) {
               PlayAvail += Read / sizeof(int16);
            }
         }

         if(Read > 0) {
            Ret = TRUE;
            if(PlayAvail == AUDIO_BUF_SIZE) {
               break;
            }
         }
         else {
            if(!feof(Playfp)) {
               LOG_ERROR(("GetAnnounceAudio(): read error : %s",
                          Err2String(errno)));
            }
            FreeAnnoucement();
         }
      }
   }

   if(PlayAvail > 0 && PlayAvail < AUDIO_BUF_SIZE) {
   // Last announcement in the chain did not fill buffer, zero the remainder
      memset(&PlayBuf[PlayAvail],0,(AUDIO_BUF_SIZE-PlayAvail) * sizeof(int16));
      PlayAvail = AUDIO_BUF_SIZE;
   }

   return Ret;
}

void QueueAudio(ConfClient *pCC,NetAudio *pAudio) 
{
   if(pAudio != NULL) {
      pCC->AudioQCount++;
      pAudio->Link = NULL;
#if 0
      if(fp != NULL) {
         fwrite(pAudio->Buf,pAudio->Available,1,fp);
      }
#endif
      if(pCC->pInHead == NULL) {
         pCC->pInHead = pAudio;
      }
      else {
         pCC->pInTail->Link = pAudio;
      }

      pCC->pInTail = pAudio;
   }
}

void DecompressGSM(ConfClient *pCC,gsm_byte *pGSM,int Len,gsm gsmh)
{
   int i;
   int GsmFrames = 0;
   NetAudio *pAudio = NULL;
   int AudioSamples = 0;
   gsm_signal *pDecoded;
   
   GsmFrames = Len / sizeof(gsm_frame);
   AudioSamples = GsmFrames * AUDIO_SAMPLES_PER_GSM_FRAME;

   if(AudioSamples > 0) {
      pAudio = (NetAudio *) malloc(sizeof(NetAudio) + 
                                   AudioSamples * sizeof(int16));
      if(pAudio == NULL) {
         LOG_ERROR(("%s#%d: malloc failed.\n",__FUNCTION__,__LINE__));
      }
      else {
         pAudio->TxMap = -1;
         pAudio->Available = AudioSamples * sizeof(int16);
         pDecoded = (gsm_signal *) pAudio->Buf;
         for(i = 0; i < GsmFrames; i++) {
            gsm_decode(gsmh,pGSM,pDecoded);
            pGSM += sizeof(gsm_frame);
            pDecoded += AUDIO_SAMPLES_PER_GSM_FRAME;
         }
         QueueAudio(pCC,pAudio);
      }
   }
}

void DecompressADPCM(ConfClient *pCC,char *p,int Len,struct adpcm_state *pState)
{
   int AudioSamples = Len * 2;
   NetAudio *pAudio = NULL;

   if(AudioSamples > 0) {
      pAudio = (NetAudio *) malloc(sizeof(NetAudio) + 
                                   AudioSamples * sizeof(int16));
      if(pAudio == NULL) {
         LOG_ERROR(("%s#%d: malloc failed.\n",__FUNCTION__,__LINE__));
      }
      else {
         pAudio->Available = AudioSamples * 2;
         adpcm_decoder(p,(short *)pAudio->Buf,AudioSamples,pState);
         QueueAudio(pCC,pAudio);
      }
   }
}

// decompress as many rtp packets as possible
void DecompressRTP(ConfClient *pCC)
{
   RtpData *pRtp = pCC->pRtpQueue;
   RtpData *pNext;
   gsm_byte *pGSM;
   int DataLen;
   char *dp;
   struct adpcm_state adpcm;      /* ADPCM compression state */
   u_int8 *sp;
   
   while(pRtp != NULL && pCC->NextSeq == pRtp->pData->seq) {
      pCC->NextSeq++;
      pCC->pRtpQueue = pRtp->Link;
      pCC->RtpQCount--;
      switch(pRtp->pData->pt) {
         case RTP_PT_GSM:
            if(pCC->gsmh == NULL) {
               pCC->gsmh = gsm_create();
            }
            pGSM = (gsm_byte *) pRtp->pData + RTP_HDR_SIZE;
            DataLen = pRtp->Len - RTP_HDR_SIZE;
            DecompressGSM(pCC,pGSM,DataLen,pCC->gsmh);
            break;

         case RTP_PT_DVI4_8K:
            sp = (u_int8 *) pRtp->pData + RTP_HDR_SIZE;
            dp = (char *) sp  + 4;
            DataLen = pRtp->Len - RTP_HDR_SIZE - 4;
            adpcm.valprev = (sp[0] << 8) | sp[1];
            adpcm.index = sp[2];
            LOG_ERROR(("Got ADPCM len: %d\n",DataLen));
            DecompressADPCM(pCC,dp,DataLen,&adpcm);
            break;

         default:
            break;
      }

      pNext = pRtp->Link;
      free(pRtp->pData);
      free(pRtp);
      pRtp = pNext;
   }
}


void DecompressAudio(ConfClient *pCC)
{
   if(pCC->Proto == PROTO_ILINK || pCC->Proto == PROTO_RTP) {
   // Currently the only supported compression technique is GSM
      DecompressRTP(pCC);
   }
   else if(pCC->Proto == PROTO_SF) {
   // Speak Freely doesn't have sequence numbers so any received data
   // has already been decompressed.  No need to do anything here.
   }
   else {
   // This shouldn't happen !
      CORE_DUMP();
   }
}


// Called when VoIP stream stops transmitting
void StopTalking(ConfClient *pCC)
{
   RtpData *pRtp;
   u_int16 NextSeqAvailable;
   
   if(pCC->pRtpQueue != NULL) {
   // flush everything that's left to audio
      while((pRtp = pCC->pRtpQueue) != NULL) {
      // There's uncompressed RTP data available
         NextSeqAvailable = pRtp->pData->seq;
         while(pCC->NextSeq != NextSeqAvailable) {
            D2PRINTF(("m1 %d",pCC->NextSeq));
            pCC->NextSeq++;
            pCC->MissedPackets++;
         }
         D2PRINTF(("\n"));
      // Make some audio
         DecompressAudio(pCC);
      }
   }

// Back to idle state
   pCC->bRxIdle = TRUE;
   pCC->bRxBuffering = pCC->bRxActive = FALSE;
}

void InsertInRTPQ(ConfClient *pCC,RtpData *pRtpData,int MaxNetQueueLen)
{
   RtpData *pLast = pCC->pRtpQueue;
   u_int16 Seq = pRtpData->pData->seq;
   
   if(Seq == ProblemSeq) {
      Breakpoint();
   }

   if(pLast == NULL) {
   // The Queue is empty
      pLast = (RtpData *) &pCC->pRtpQueue;
   }
   else {
      RtpData *p = pLast->Link;

      if(p == NULL) {
      // only one packet on the queue
         if(Seq == pLast->pData->seq) {
            pCC->DuplicatePackets++;
            free(pRtpData->pData);
            free(pRtpData);
            pLast = NULL;
         }
         else if(Between(Seq,pLast->pData->seq,pLast->pData->seq + MaxNetQueueLen)) {
         // this packet is now the first on the queue
            D2PRINTF(("\ns(%d,%d)",pLast->pData->seq,Seq));
            pCC->OutofSeqPackets++;
            pRtpData->Link = pCC->pRtpQueue;
            pLast = (RtpData *) &pCC->pRtpQueue;
         }
      }
      else  {
         while(p != NULL) {
            if(Seq == pLast->pData->seq) {
               pCC->DuplicatePackets++;
               free(pRtpData->pData);
               free(pRtpData);
               pLast = NULL;
               break;
            }
            else if(Between(pLast->pData->seq,Seq,p->pData->seq)) {
            // insert new packet here
               pRtpData->Link = p;
               pCC->OutofSeqPackets++;
               D2PRINTF(("\ns(%d,%d,%d)",pLast->pData->seq,Seq,p->pData->seq));
               break;
            }
            pLast = p;
            p = p->Link;
         }
      }
   }

   if(pLast != NULL) {
      D2PRINTF(("q"));
      pLast->Link = pRtpData;
      pCC->RtpQCount++;
   }
}

void CPort::MixAudio(int16 *In)
{
   int *Out = MixBuffer;
   int Sample;
   int i;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      *Out += *In++;
      Sample = *Out++;
      if(Sample < 0) {
         Sample = -Sample;
      }

      if(Sample > MaxMixedLevel) {
         MaxMixedLevel = Sample;
      }
   }
}

// Scale the mixed audio so that the maximum level does not exceed 
// MAX_MIXED_LEVEL.  When the mix gain changes try to change the gain
// at a zero crossing to avoid thumps.
void CPort::ScaleMixedAudio()
{
   int MixGain;
   int i;
   int bFindNegCrossing = MixBuffer[0] > 0 ? TRUE : FALSE;

   if(MaxMixedLevel > MAX_MIXED_LEVEL) {
   // calculate gain to reduce maximum mixed level to 
      MixGain = (MAX_MIXED_LEVEL * 1024) / MaxMixedLevel;
   }
   else {
      MixGain = 1024;   // 1:1
   }

   if(MixGain == 1024 && LastMixGain == 1024) {
   // Both this gain and the last gain were 1:1, so no need to multiply
      for(i = 0; i < AUDIO_BUF_SIZE; i++) {
         AudioBuf[i] = (int16) MixBuffer[i];
      }
   }
   else {
      if((MixGain - LastMixGain) > 2) {
      // increase gain slowly to make it less noticeable
         MixGain = LastMixGain + (MixGain - LastMixGain)/2;
      }

      for(i = 0; i < AUDIO_BUF_SIZE; i++) {
         if(MixBuffer[i] > MAX_MIXED_LEVEL || MixBuffer[i] < -MAX_MIXED_LEVEL) {
         // need to apply new MixGain now
            break;
         }

         if(bFindNegCrossing) {
            if(MixBuffer[i] < 0) {
            // apply new gain now
               break;
            }
         }
         else {
            if(MixBuffer[i] > 0) {
            // apply new gain now
               break;
            }
         }

      // AudioBuf[i] = MixBuffer[i] * LastMixGain / 1024;
         AudioBuf[i] = (int16) ((MixBuffer[i] * LastMixGain) >> 10);
      }

   // change to the new mix gain for the remainder of the buffer
      LastMixGain = MixGain;

      for(; i < AUDIO_BUF_SIZE; i++) {
         AudioBuf[i] = (int16) ((MixBuffer[i] * LastMixGain) >> 10);
      }
   }
}

// Fill AudioBuf with client audio
int CPort::GetClientAudio(ConfClient *pCC)
{
   int CopyCount;
   RtpData *pRtp;
   u_int16 NextSeqAvailable;
   NetAudio *pAudio;
   PCMAudio *pPcm;
   PCMAudio *pLastPcm = (PCMAudio *) &pCC->pPcmHead;
   int Duration;
   long microseconds;

   AudioAvailable = 0;

   while(AudioAvailable < sizeof(AudioBuf)) {
   // need more audio in the output buffer
      pAudio = NULL;
      pPcm = pCC->pPcmHead;

      while(pPcm != NULL) {
         if(pPcm->Port == PortNum) {
            pAudio = pPcm->pNetAudio;
            break;
         }
         pLastPcm = pPcm;
         pPcm = pPcm->Link;
      }

      if(pAudio == NULL) {
      // Try to find a new NetAudio buffer for us
         pAudio = pCC->pInHead;
         while(pAudio != NULL) {
            if(pAudio->TxMap & MyBit) {
               break;
            }
            pAudio = pAudio->Link;
         }

         if(pAudio != NULL) {
            if((pPcm = (PCMAudio *) malloc(sizeof(PCMAudio))) == NULL) {
               LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            }
            else {
               pPcm->pNetAudio = pCC->pInHead;
               pPcm->Port = PortNum;
               pPcm->Used = 0;
               pPcm->Link = pCC->pPcmHead;
               pCC->pPcmHead = pPcm;
               pAudio = pPcm->pNetAudio;
            }
         }
      }

      if(pAudio != NULL) {
         CopyCount = sizeof(AudioBuf) - AudioAvailable;
         CopyCount = min(CopyCount,pAudio->Available - pPcm->Used);
         D2PRINTF(("Copying %d bytes to AudioBuf[%d]\n",CopyCount,AudioAvailable));
         memcpy(&AudioBuf[AudioAvailable/2],&pAudio->Buf[pPcm->Used],CopyCount);
         AudioAvailable += CopyCount;
         pPcm->Used += CopyCount;
         if(pPcm->Used == pAudio->Available) {
         // Finished with this buffer
            pLastPcm->Link = pPcm->Link;
            free(pPcm);
            pAudio->TxMap &= ~MyBit;
            if((pCC->TxMap & pAudio->TxMap) == 0) {
               D2PRINTF(("Buffer emptied\n"));
               pCC->pInHead = pAudio->Link;
               pCC->AudioQCount--;
               free(pAudio);
            }
         }
      }
      else if(pCC->bRxActive && (pRtp = pCC->pRtpQueue) != NULL) {
      // We're not buffering and there's uncompressed RTP data available
         NextSeqAvailable = pRtp->pData->seq;
         while(pCC->NextSeq != NextSeqAvailable) {
            D2PRINTF(("m2 %d ",pCC->NextSeq));
            pCC->NextSeq++;
            pCC->MissedPackets++;
         }
         D2PRINTF(("\n"));
      // Try to decompress some audio
         DecompressAudio(pCC);
      }
      else if(pCC->bRxActive) {
      // Nothing more we can do, we're out out data.  
      // Go back to buffering state and get some more data
         pCC->bRxActive = FALSE;
         pCC->bRxBuffering = TRUE;
         D2PRINTF(("O"));
         break;
      }
      else {
         break;
      }
   }

   if(AudioAvailable > 0 && AudioAvailable < sizeof(AudioBuf)) {
   // Got some audio, but not enough. Zero the remainder of the buffer
      D2PRINTF(("\nZeroing %d bytes at end of AudioBuf\n",sizeof(AudioBuf)-AudioAvailable));
      memset(&AudioBuf[AudioAvailable/2],0,sizeof(AudioBuf)-AudioAvailable);
      AudioAvailable = sizeof(AudioBuf);
#if 0
      if(fp != NULL && PortNum == 0) {
         fwrite(AudioBuf,AudioAvailable,1,fp);
      }
#endif
   }

   if(AudioAvailable > 0 && VoipCtone != 0 && (VoipCtoneMap & MyBit) && 
      !bTimeOut) 
   {
   // (re)start the VoipCtoneTimer to go off after the audio 
   // we're about to output has been consumed.
      if(VoipCtoneTimer.tv_sec == 0) {
      // Timer is not running, start it
         VoipCtoneTimer = TimeNow;
      }
      Duration = (AudioAvailable * 500) / 8000; // in milliseconds

      microseconds = VoipCtoneTimer.tv_usec + ((Duration % 1000) * 1000);
      VoipCtoneTimer.tv_usec = microseconds % 1000000;
      VoipCtoneTimer.tv_sec  = VoipCtoneTimer.tv_sec + (Duration / 1000) + 
                               microseconds / 1000000;
   }
   return AudioAvailable;
}

// Count the number of ports that have audio ready for this port:
//    0 - none
//    1 - exactly one
//    2 - 2 or more
int CPort::SourcesReady(ConfClient **pTalker)
{
   int Ret = 0;
   struct avl_traverser avl_trans;
   ConfClient *pCC;

   if(RxMap & CosMap) {
   // At least one the ports we're listening to is active
      int Port;
      int Bits = RxMap & CosMap;

      for(Port = 0; Port < NumPorts; Port++) {
         if(Bits & 1) {
         // We're listening to this port and it's active
            if(GetPortPtr(Port)->RxAudioReady & MyBit) {
            // Port has audio ready for us
               if(Ret++ == 1) {
               // 2 active ports, no need to count more
                  break;
               }
            }
         }
         Bits >>= 1;
         if(Bits == 0) {
         // No more ports to test
            break;
         }
      }
   }

   if(Ret < 2) {
   // 1 or few sources, add number of VoIP connections until we have
   // 2 or more (enough to tell if we need to mix audio or not)
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(VoipAudioAvailable(pCC)) {
            if(pTalker != NULL) {
               *pTalker = pCC;
            }
            if(Ret++ >= 2) {
            // 2 or more active ports, no need to look further
               break;
            }
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
   }

   return Ret;
}

// Return TRUE if there are active ports.  
// NB: This does *NOT* mean that a port has have audio ready
// currently.  i.e. VoIP clients may be in the buffering state.
int CPort::ActivePorts()
{
   int Ret = FALSE;
   struct avl_traverser avl_trans;
   ConfClient *pCC;

   do {
      if(RxMap & CosMap) {
      // At least one the ports we're listening to is active
         Ret = TRUE;
         break;
      }

      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if((pCC->TxMap & MyBit) && !pCC->bRxIdle) {
            Ret = TRUE;
            break;
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
   } while(FALSE);

   return Ret;
}

void CPort::FillAudioBuf()
{
   ConfClient *pCC;
   ConfClient *pTalker = NULL;
   struct avl_traverser avl_trans;
   int Sources = 0;
   ConfClient *pPlayer = piLinkConf->pFilePlayer;
   CPort *pPort;
   CToneGen *pToneGen;
   CToneGen *pLastGen;
   CToneGen *pNextGen;

// Count the sources that are talking to see if we need to mix or not
   if(GetAnnounceAudio()) {
      Sources++;
   }

   if(CWChar != NULL) {
      Sources++;
   }

   Sources += SourcesReady(&pTalker);

   pToneGen = ToneGenerators;
   while(pToneGen != NULL) {
      Sources++;
      pToneGen = pToneGen->Link;
   }

   if(pPlayer != NULL && (!pPlayer->bRxIdle || pPlayer->pInHead)) {
      Sources++;
   }

   if(Sources > 0 && CTCSSLevel > 0) {
      Sources++;
   }

   if(Sources > 1 || LastMixGain != 1024) {
   // more than one source is active or the last mix gain wasn't 1:1, mix them
      memset(MixBuffer,0,sizeof(MixBuffer));
      MaxMixedLevel = 0;
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(pCC->TxMap & MyBit && !pCC->bRxIdle || pCC->pInHead != NULL) {
            if(GetClientAudio(pCC)) {
               MixAudio(AudioBuf);
            }
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }

      if(pPlayer != NULL && (!pPlayer->bRxIdle || pPlayer->pInHead)) {
         if(GetClientAudio(pPlayer)) {
            MixAudio(AudioBuf);
         }
      }

      if(PlayAvail > 0) {
         MixAudio(PlayBuf);
      }

      if(CWChar != NULL) {
         GenCW(AudioBuf);
         MixAudio(AudioBuf);
      }

      pToneGen = ToneGenerators;
      pLastGen = NULL;
      while(pToneGen != NULL) {
         pNextGen = pToneGen->Link;
         if(pToneGen->ToneGen(AudioBuf,AUDIO_BUF_SIZE)) {
         // This generator has finished
            if(pLastGen == NULL) {
               ToneGenerators = pNextGen;
            }
            else {
               pLastGen->Link = pNextGen;
            }
            delete pToneGen;
         }
         else {
            pLastGen = pToneGen;
         }
         MixAudio(AudioBuf);
         pToneGen = pNextGen;
      }

      if(CTCSSLevel > 0) {
         GenCTCSS(AudioBuf);
         MixAudio(AudioBuf);
      }

      if(RxMap != 0) {
         int Port;
         int Bits = RxMap;

         for(Port = 0; Port < MAX_PORTS; Port++) {
            if(Bits & 1) {
            // We're listening to this port
               pPort = GetPortPtr(Port);
               if(pPort->RxAudioReady & MyBit) {
               // Port has audio ready for us
                  MixAudio(pPort->AudioInBuf);
                  pPort->RxAudioReady &= ~MyBit;
               }
            }
            Bits >>= 1;
            if(Bits == 0) {
            // No more ports to test
               break;
            }
         }
      }

   // Now apply the mix gain to prevent over modulation

      ScaleMixedAudio();
      AudioAvailable = sizeof(AudioBuf);
   }
   else if(Sources == 1) {
   // Simple optimization, just a single source so there's no need to mix
      AudioAvailable = sizeof(AudioBuf);
      if(pTalker != NULL) {
         GetClientAudio(pTalker);
      }
      else if(PlayAvail > 0) {
         memcpy(AudioBuf,PlayBuf,PlayAvail*sizeof(int16));
      }
      else if(CWChar != NULL) {
         GenCW(AudioBuf);
      }
      else if(ToneGenerators != NULL) {
         if(ToneGenerators->ToneGen(AudioBuf,AUDIO_BUF_SIZE)) {
         // This generator has finished
            D2PRINTF(("%s#%d: Deleting tonegen #%d\n",__FUNCTION__,__LINE__,
                      ToneGenerators->ID));
            delete ToneGenerators;
            ToneGenerators = NULL;
         }
      }
      else if(CTCSSLevel > 0) {
         GenCTCSS(AudioBuf);
      }
      else if(pPlayer != NULL && (!pPlayer->bRxIdle || pPlayer->pInHead)) {
         GetClientAudio(pPlayer);
      }
      else if(RxMap != 0) {
         int Port;
         int Bits = RxMap & CosMap;

         for(Port = 0; Port < MAX_PORTS; Port++) {
            if(Bits & 1) {
            // We're listening to this port
               pPort = GetPortPtr(Port);
               if(pPort->RxAudioReady & MyBit) {
               // Port has audio ready for us
                  memcpy(AudioBuf,pPort->AudioInBuf,sizeof(AudioBuf));
                  pPort->RxAudioReady &= ~MyBit;
               // No point in scanning more ports, we already know Sources == 1
                  break;
               }
            }
            Bits >>= 1;
         }
      }
      else {
      // this shouldn't happen
         CORE_DUMP();
      }
   }
   else {
   // no audio sources
      LastMixGain = 1024;
      AudioAvailable = 0;
   }

   if(bTimeOut) {
   // Only return audio from the Timeout tone if timed out
      if(TimeoutToneGen != NULL) {
         AudioAvailable = AUDIO_BUF_SIZE;
         if(TimeoutToneGen->ToneGen(AudioBuf,AUDIO_BUF_SIZE)) {
         // This generator has finished
            D2PRINTF(("%s: Deleting timeout tonegen #%d\n",__FUNCTION__,
                      TimeoutToneGen->ID));
            delete TimeoutToneGen;
            TimeoutToneGen = NULL;
         }
      }
      else {
         AudioAvailable = 0;
      }
   }
}

// return TRUE if we write audio to the sound device
int CPort::WriteAudioOut(ClientInfo *p)
{
   int wrote;
   char *pWriteData;
   int Ret = FALSE;

   bAudioOutCalled = TRUE;
   
   for(; ; ) {
      if(AudioAvailable == 0) {
         AudioUsed = 0;
         D2PRINTF(("F"));
         FillAudioBuf();
      }

      if(AudioAvailable > 0) {
         pWriteData = ((char *) AudioBuf) + AudioUsed;
         wrote = write(p->Socket,pWriteData,AudioAvailable);
#if 0
         if(fp != NULL && wrote > 0 && PortNum == 0) {
            fwrite(pWriteData,wrote,1,fp);
         }
#endif
         if(wrote <= 0) {
            if(errno == EAGAIN) {
               WriteWaits++;
               p->bWriteWait = TRUE;
               D2PRINTF(("W"));
            }
            else {
               LOG_ERROR(("WriteAudioOut(): write() failed: %s",
                          Err2String(errno)));
            }
            break;
         }
         else {
            D2PRINTF(("w"));
            HangTimer.tv_sec = 0;
            if(!bTxKeyed) {
               KeyTx(TRUE);
            }
            AudioUsed += wrote;
            AudioAvailable -= wrote;
            Ret = TRUE;
         }
      }
      else {
      // No more audio available now
         WriteStalls++;
         p->bWriteWait = FALSE;
         break;
      }
   }
   DFLUSH();

   return Ret;
}

//
// RxState  
void QueueRTPData(ConfClient *pCC,RtpData *pRtp)
{
   u_int16 Seq = pRtp->pData->seq;
   int NetQueueLen = 0;
   int MaxNetQueueLen;

   switch(pRtp->pData->pt) {
      case RTP_PT_GSM:
         NetQueueLen = VoipDejitterDelay / 80;
         break;

      case RTP_PT_DVI4_8K:
         NetQueueLen = VoipDejitterDelay / 122;
         break;

      default:
         break;
   }
   MaxNetQueueLen = NetQueueLen * 2;

   if(pCC->bRxIdle) {
   // First packet received, assume this is the first sequence #
      D2PRINTF(("t%d",Seq));
      pCC->NextSeq = Seq;
      pCC->bRxIdle = FALSE;
      pCC->bRxBuffering = TRUE;
   }

   if(pCC->bRxBuffering) {
      if(Between(Seq,pCC->NextSeq,Seq + MaxNetQueueLen)) {
      // This packet is before the current NextSeq, backup NextSeq
         pCC->NextSeq = Seq;
      }
   }

   if(Between(pCC->NextSeq,Seq,pCC->NextSeq + MaxNetQueueLen)) {
      InsertInRTPQ(pCC,pRtp,MaxNetQueueLen);
      if(pCC->bRxBuffering) {
         if(pCC->RtpQCount >= NetQueueLen) {
         // We're buffering and it's time to rock and roll
            D2PRINTF(("S"));
            pCC->bRxBuffering = FALSE;
            pCC->bRxActive = TRUE;
            DecompressAudio(pCC);
         }
      }
      else if(pCC->bRxActive) {
      // try to make some more audio
         DecompressAudio(pCC);
      }
      else {
      // this shouldn't happen
         CORE_DUMP();
      }
   }
   else {
   // This packet is out of range, must have been delayed too long,
   // toss it
      D2PRINTF(("\nl(%d,%d,%d)",pCC->NextSeq,Seq,pCC->NextSeq+MaxNetQueueLen));
      pCC->LatePackets++;
      free(pRtp->pData);
      free(pRtp);
   }
}

void QueueSFData(ConfClient *pCC,soundbuf *p)
{
   int compression = (int) htonl(p->compression);
   char *pData;
   int Len = htonl(p->buffer.buffer_len);
   struct adpcm_state adpcm;      /* ADPCM compression state */
   u_int8 *sp;
   int NetQueueLen = 0;

   if(pCC->bRxIdle) {
   // First packet received, assume this is the first sequence #
      D2PRINTF(("t"));
      pCC->bRxIdle = FALSE;
      pCC->bRxBuffering = TRUE;
   }

// Speak Freely doesn't have sequence numbers so decompress data now
   switch(compression) {
      case fCompGSM | fProtocol:
         if(pCC->gsmh == NULL) {
            pCC->gsmh = gsm_create();
         }
         Len -= 2;
         pData = &p->buffer.buffer_val[2];
         NetQueueLen = VoipDejitterDelay / 200;
         DecompressGSM(pCC,(gsm_byte *) pData,Len,pCC->gsmh);
         break;

      case fCompADPCM | fProtocol:
       // Restore the decoder state from the copy saved in the 
       // packet, independent of the byte order of the machine 
       // we're running on. 
         pData = p->buffer.buffer_val;
         Len -= 3;
         sp = (u_int8 *) &pData[Len];
         adpcm.valprev = (sp[0] << 8) | sp[1];
         adpcm.index = sp[2];
         NetQueueLen = VoipDejitterDelay / 122;
         DecompressADPCM(pCC,pData,Len,&adpcm);
         break;

      default:
         CORE_DUMP();
         break;
   }

   if(pCC->bRxBuffering) {
      if(pCC->AudioQCount == NetQueueLen) {
      // We're buffering and it's time to rock and roll
         D2PRINTF(("S"));
         pCC->bRxBuffering = FALSE;
         pCC->bRxActive = TRUE;
      }
      else {
         D2PRINTF(("q"));
      }
   }
   else if(!pCC->bRxActive) {
   // this shouldn't happen
      CORE_DUMP();
   }
}


void EndPointCleanup(ConfClient *pCC)
{
   NetAudio *pAudio;
   RtpData *pRtp;

   if(pCC->gsmh != NULL) {
      gsm_destroy(pCC->gsmh);
   }

   while((pAudio = pCC->pInHead) != NULL) {
      pCC->pInHead = pAudio->Link;
      pCC->AudioQCount--;
      free(pAudio);
   }

   while((pRtp = pCC->pRtpQueue) != NULL) {
      pCC->pRtpQueue = pRtp->Link;
      free(pRtp->pData);
      free(pRtp);
   }
}

// Called by "conference" code when an rtp event occurs
int EndPoint(ClientInfo *p,ConfClient *pCC,EventType Event)
{
   RtpData *pRtpData;

   switch(Event) {
      case EVENT_INIT:
         pCC->bRxIdle = TRUE;
         pCC->RxMap = VoipInMap;
         pCC->TxMap = VoipInMap;
         break;

      case EVENT_RTP_RX:
        if(pCC->bInConf && !pCC->bMuted && pCC->bTalking) {
            if(pCC->Proto == PROTO_ILINK || pCC->Proto == PROTO_RTP) {
               pRtpData = (RtpData *) malloc(sizeof(RtpData));
               if(pRtpData != NULL) {
                  pRtpData->pData = (rtp_hdr_t *) p->Buf;
                  pRtpData->Len = p->Count;
                  pRtpData->Link = NULL;
                  p->Buf = (char *) malloc(p->BufSize);
               // For debug convenance convert the sequence number into host order
                  pRtpData->pData->seq = ntohs(pRtpData->pData->seq);
               }
               QueueRTPData(pCC,pRtpData);
               DFLUSH();
            }
            else if(pCC->Proto == PROTO_SF) {
               QueueSFData(pCC,(soundbuf *) p->Buf);
            }
            else {
            // This shouldn't happen !
               CORE_DUMP();
            }
         }
         break;

      case EVENT_RTP_TO:
         D2PRINTF(("e"));
         StopTalking(pCC);
         DFLUSH();
         break;

      case EVENT_RTCP_RX:
      case EVENT_RTCP_TO:
         break;

      case EVENT_DELETE:
      // cleanup, this guy is going to be history shortly.
#if 0
         if(fp != NULL) {
            fclose(fp);
            fp = NULL;
         }
#endif
         EndPointCleanup(pCC);
         break;
   }
   return FALSE;
}

void ExpandCommand(char *Command,char *DtmfCommand,char *DtmfWildCards)
{
   char *cp;
   char *cp1;
   int ExtraSpace = 0;
   int WildCardChars = strlen(DtmfWildCards);
   int DtmfChars = strlen(DtmfCommand);

// Count the space that will be needed when the command is expanded
   cp = Command;
   while((cp = strchr(cp,'$')) != NULL) {
      cp++;
      switch(*cp) {
         case '1':
         // We'll replace "$1" with DtmfCommand
            ExtraSpace -= 2;
            ExtraSpace += DtmfChars;
            break;

         case '2':
         // We'll replace "$2" with DtmfWildCards
            ExtraSpace -= 2;
            ExtraSpace += WildCardChars;
            break;
#if 0
         case '3':
         // We'll replace "$3" with the current port number
            ExtraSpace -= 2;
            ExtraSpace += PortNum > 9 ? 3 : 2;
            break;

         case '$':
         // 
            break;
#endif

         default:
            break;
      }
   }

   if(ExtraSpace != 0) {
   // Something needs expanding
      char *Temp = (char *) malloc(strlen(Command) + ExtraSpace + 1);
      if(Temp != NULL) {
         cp = Command;
         cp1 = Temp;
         while(*cp) {
            if(*cp == '$' && cp[1] == '1') {
            // replace "$1" with DtmfCommand
               strcpy(cp1,DtmfCommand);
               cp1 += DtmfChars;
               cp += 2;
            }
            else if(*cp == '$' && cp[1] == '2') {
            // Replace "$2" with DtmfWildCards
               strcpy(cp1,DtmfWildCards);
               cp1 += WildCardChars;
               cp += 2;
            }
            else {
            // Just copy the character
               *cp1++ = *cp++;
            }
         }

         *cp1 = 0;
      // Execute the expanded command
         ExecuteCommand(Temp);
         free(Temp);
      }
      else {
         LOG_ERROR(("%s#%d: malloc failed.\n",__FUNCTION__,__LINE__));
      }
   }
   else {
   // Nothing to expand, just execute the command as is
      ExecuteCommand(Command);
   }
}

void CPort::DecodeDTMF(char *Command)
{
   DTMF_Command *p = pDTMFCommands;
   char  WildCardBuf[MAX_DTMF_DIGITS+1];
   int i;
   int j;
   int WildCardCount = 0;
   char *cp;
   char *CmdPtr;
   int Digits = strlen(Command);

   LOG_NORM(("%s: Decoding \"%s\"\n",__FUNCTION__,Command));

   while(p != NULL) {
      for(i = 0, j = 0; i < Digits; i++) {
         if(p->DTMF_String[j] == '?' || p->DTMF_String[j] == '+') {
         // Wildcard match, save the matched character
            WildCardBuf[WildCardCount++] = Command[i];
            if(p->DTMF_String[j] == '?') {
               j++;
            }
         }
         else if(p->DTMF_String[j] == Command[i]) {
         // match
            j++;
         }
         else {
         // no match
            break;
         }
      }

      if(i == Digits && (p->DTMF_String[j] == 0 || p->DTMF_String[j] == '+')) {
      // Found a matching command
         break;
      }
   // No match, keep looking
      p = p->Link;
   }

   if(p != NULL) {
   // Found the command, execute it
      char *Temp;
      WildCardBuf[WildCardCount] = 0;

   // Make a copy of the command in case the command causes a new
   // command file to be loaded (which will free the passed in string).
      Temp = strdup(p->Command);
      CmdPtr = Temp;
      while(CmdPtr != NULL) {
         if((cp = strchr(CmdPtr,';')) != NULL) {
            *cp = 0;
         }
      // Expand and execute the command
         ExpandCommand(CmdPtr,Command,WildCardBuf);

         if(cp != NULL) {
         // restore the command seperator character we removed earlier
            *cp++ = ';';
         // Skip any white space to make the logs look pretty
            while(isspace(*cp)) {
               cp++;
            }
         }

         CmdPtr = cp;
      }
      free(Temp);
   }
   else {
      LOG_ERROR(("Failed to decode touchtone command \"%s\"\n",Command));
   }
}

void CPort::StartTalking(int bTalk)
{
   if(bTalking != bTalk) {
      bTalking = bTalk;
      if(!bTalk && DTMF_Digits > 0) {
         DecodeDTMF(DTMF_Buf);
         DTMF_Digits = 0;
      }

      if(!bTalk) {
         LastSpoke = TimeNow;
      }
   }
}

void CPort::Vox()
{
   int AveLevel = 0;
   int i;
   int Samples = AudioInAvailable / 2;

   for(i = 0; i < Samples; i++) {
      AveLevel += abs(AudioInBuf[i]);
   }

   AveLevel /= Samples;
   
   if(AveLevel > VoxThreshold) {
      if(!bTalking) {
         StartTalking(TRUE);
         D2PRINTF(("V"));
         DFLUSH();
      }
      LastVoxTrip = TimeNow;
   }
   else if(bTalking && TimeLapse(&LastVoxTrip) > VoxHoldTime) {
      StartTalking(FALSE);
      D2PRINTF(("v"));
      DFLUSH();
   }
}

char tone_detector_input(int16 *s, int n);

void CPort::AdpcmTx(int SfAdpcmCount,int RtpAdpcmCount)
{
   int Samples2Encode;
   int SamplesUsed = 0;
   ConfClient *pCC;
   struct avl_traverser avl_trans;

   if(SfAdpcmCount > 0 || RtpAdpcmCount > 0) do {
      Samples2Encode = (ADPCM_PACKET_DATA_BYTES - ADPCM_Bytes) * 2;
      if(Samples2Encode > (AUDIO_BUF_SIZE - SamplesUsed)) {
         Samples2Encode = AUDIO_BUF_SIZE - SamplesUsed;
      }
#if 0
      printf("Encode %d samples from %d to %d\n",
             Samples2Encode,SamplesUsed,ADPCM_Bytes);
#endif
      adpcm_coder(&AudioInBuf[SamplesUsed],&Adpcm.Buf[ADPCM_Bytes],
                  Samples2Encode,&TxADPCM);
      ADPCM_Bytes += Samples2Encode/2;
      SamplesUsed += Samples2Encode;

      if(ADPCM_Bytes == ADPCM_PACKET_DATA_BYTES) {
      // The ADPCM packet is full, sendit
         ADPCM_Bytes = 0;

//       printf("packet full, sendit\n");

         if(SfAdpcmCount > 0) {
         // Fill in the Send Speak Freely header
            soundbuf *pSFHdr = (soundbuf *) &Adpcm.Buf[-SIZEOF_SOUNDBUF];
            u_int8 *pData = (u_int8 *) Adpcm.Buf;
            u_int8 *sp = (u_int8 *) &pData[ADPCM_PACKET_DATA_BYTES];
            int TxLen = SIZEOF_SOUNDBUF + ADPCM_PACKET_DATA_BYTES + 3;

            pSFHdr->compression = (int) htonl(fCompADPCM| fProtocol);
            pSFHdr->buffer.buffer_len = (int) htonl(ADPCM_PACKET_DATA_BYTES+3);
            memset(pSFHdr->sendinghost,0,sizeof(pSFHdr->sendinghost));
            strncpy(pSFHdr->sendinghost,ConferenceCall,
                    sizeof(pSFHdr->sendinghost));
            sp[0] = (u_int8) (((u_int16)TxADPCMStart.valprev) >> 8);
            sp[1] = (TxADPCMStart.valprev & 0xff);
            sp[2] = TxADPCMStart.index;

         // Send Speak Freely ADPCM packets
            pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
            while(pCC != NULL) {
               if(pCC->RxMap & MyBit && 
                  pCC->bInConf && 
                  pCC->CompressionType == RTP_PT_DVI4_8K &&
                  pCC->Proto == PROTO_SF)
               {
                  Send2(pCC,pSFHdr,TxLen,pCC->pCS->AudioPort);
               }
               pCC = (ConfClient *) avl_t_next(&avl_trans);
            }
         }

         if(RtpAdpcmCount > 0) {
         // Fill in the RTP header
            rtp_hdr_t *pRtpHdr = (rtp_hdr_t *) &Adpcm.Buf[-(RTP_HDR_SIZE+4)];
            u_int8 *sp = (u_int8 *) &Adpcm.Buf[-4];
            int TxLen = RTP_HDR_SIZE + ADPCM_PACKET_DATA_BYTES + 4;

            memset(pRtpHdr,0,sizeof(rtp_hdr_t));
            pRtpHdr->pt = RTP_PT_DVI4_8K;
            pRtpHdr->seq = htons(RtpTxSequence++);
            pRtpHdr->ssrc = OurNodeID;
            pRtpHdr->version = 2;
            sp[0] = (u_int8) (((u_int16)TxADPCMStart.valprev) >> 8);
            sp[1] = (TxADPCMStart.valprev & 0xff);
            sp[2] = TxADPCMStart.index;

         // Send RTP ADPCM packets
            pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
            while(pCC != NULL) {
               if(pCC->RxMap & MyBit && pCC->bInConf && 
                  pCC->CompressionType == RTP_PT_DVI4_8K &&
                  pCC->Proto == PROTO_RTP)
               {
                  Send2(pCC,pRtpHdr,TxLen,pCC->pCS->AudioPort);
               }
               pCC = (ConfClient *) avl_t_next(&avl_trans);
            }
         }
         TxADPCMStart = TxADPCM;
      }
   } while(SamplesUsed < AUDIO_BUF_SIZE);
}

void CPort::GsmTx(int SfGsmCount,int RtpGsmCount,int EchoLinkCount)
{
   ConfClient *pCC;
   struct avl_traverser avl_trans;
   gsm_byte *pGSM;
   int i;
   rtp_hdr_t *pRtpHdr;
   char SaveBuf[sizeof(rtp_hdr_t)];

   pGSM = (gsm_byte *) &GSM_Packet[SIZEOF_SOUNDBUF+2];
   pGSM += GSM_FRAME_LEN * GsmFramesAvail;

// Compress the audio

   for(i = 0; i < AUDIO_BUF_SIZE; i += AUDIO_SAMPLES_PER_GSM_FRAME) {
      gsm_encode(hGSM_Encode,&AudioInBuf[i],pGSM);
      pGSM += sizeof(gsm_frame);
   }
   pGSM = (gsm_byte *) &GSM_Packet[SIZEOF_SOUNDBUF+2];
   pGSM += GSM_FRAME_LEN * GsmFramesAvail;


   if(RtpGsmCount > 0 || EchoLinkCount > 0) {
   // Send Echolink and RTP packets

      pRtpHdr = (rtp_hdr_t *) (((char *) pGSM) - RTP_HDR_SIZE);
      if(GsmFramesAvail > 0) {
      // Save the data we're about to clobber locally
         memcpy(SaveBuf,pRtpHdr,sizeof(rtp_hdr_t));
      }
      memset(pRtpHdr,0,RTP_HDR_SIZE);
      pRtpHdr->pt = RTP_PT_GSM;
      pRtpHdr->seq = htons(TxSequence++);
      pRtpHdr->ssrc = OurNodeID;

      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(pCC->RxMap & MyBit && 
            pCC->bInConf && pCC->CompressionType == RTP_PT_GSM) 
         {
            if(pCC->Proto == PROTO_ILINK) {
               pRtpHdr->version = ILINK_RTP_VERSION;
               Send2(pCC,pRtpHdr,ILINK_RTP_FRAME_LEN,ILINK_RTP_PORT);
            }
            else if(pCC->Proto == PROTO_RTP) {
               pRtpHdr->version = 2;
               Send2(pCC,pRtpHdr,ILINK_RTP_FRAME_LEN,pCC->pCS->AudioPort);
            }
            else {
               CORE_DUMP();
            }
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }

      if(GsmFramesAvail > 0) {
      // Restore the data we've clobbered
         memcpy(pRtpHdr,SaveBuf,sizeof(rtp_hdr_t));
      }
   }

   if(SfGsmCount > 0) do {
   // Fill in the Send Speak Freely header
      soundbuf *pSFHdr = (soundbuf *) GSM_Packet;
      int TxLen;

      GsmFramesAvail += AUDIO_BUF_SIZE / AUDIO_SAMPLES_PER_GSM_FRAME;
      if(GsmFramesAvail < 10) {
      // Not enough data yet
         break;
      }
      GsmFramesAvail -= 10;

   // Hide the original number of samples in first 2 bytes of buffer.
      *((short *) pSFHdr->buffer.buffer_val) = htons(AUDIO_SAMPLES_PER_GSM_FRAME*10);

      pSFHdr->compression = (int) htonl(fCompGSM | fProtocol);
      pSFHdr->buffer.buffer_len = (int) htonl(2 + (10*GSM_FRAME_LEN));
      TxLen = SIZEOF_SOUNDBUF + 2 + (10*GSM_FRAME_LEN);

      memset(pSFHdr->sendinghost,0,sizeof(pSFHdr->sendinghost));
      strncpy(pSFHdr->sendinghost,ConferenceCall,sizeof(pSFHdr->sendinghost));

   // Send Speak Freely ADPCM packets

      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(pCC->RxMap & MyBit && pCC->bInConf && 
            pCC->CompressionType == RTP_PT_GSM &&
            pCC->Proto == PROTO_SF) 
         {  
            Send2(pCC,pSFHdr,TxLen,pCC->pCS->AudioPort);
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }

      if(GsmFramesAvail > 0) {
      // Copy GSM data we didn't use to start of buffer
         char *pTo = &GSM_Packet[SIZEOF_SOUNDBUF+2];
         char *pFrom = pTo + (GSM_FRAME_LEN * 10);
         memcpy(pTo,pFrom,GsmFramesAvail * GSM_FRAME_LEN);
      }
   } while(FALSE);
}

void CPort::ProcessAudioIn(ClientInfo *p)
{
   ConfClient *pCC;
   struct avl_traverser avl_trans;
   int SfAdpcmCount = 0;
   int RtpAdpcmCount = 0;
   int SfGsmCount = 0;
   int RtpGsmCount = 0;
   int EchoLinkCount = 0;
   int GsmCount = 0;
   int AdpcmCount = 0;

   if(HalfDuplexRF && bTxKeyed) {
      StartTalking(FALSE);
   }
   else {
      if(RxCosMethod == 1) {
         Vox();
      }
   }
   
   if(DecodeCTCSS) {
      int HadCTCSS = g_bRxCTCSS;

      UpdateCTCSS(AudioInBuf,AUDIO_BUF_SIZE);
      if(HadCTCSS != g_bRxCTCSS) {
      // CTCSS decode state has changed
         DPRINTF(("%s: Port %s CTCSS decode %sactive\n",__FUNCTION__,PortName,
                 g_bRxCTCSS ? "" : "in"));
         StartTalking(g_bRxCTCSS && g_bCosActive);
      }
   }

   if(bTalking) {
#if 0
      if(fp != NULL) {
         fwrite(AudioInBuf,sizeof(AudioInBuf),1,fp);
      }
#endif

      if(DtmfMethod == 3) {
         dtmf_rx(&DtmfRxState,AudioInBuf,AUDIO_BUF_SIZE);
      }
   }

   if(DTMF_Digits > 0 && TimeLapse(&LastDTMFDigit) > DTMFCommandTimeout) {
      DecodeDTMF(DTMF_Buf);
      DTMF_Digits = 0;
   }

   if(!bTalking && bUngatedRx && !bTossRxIn) {
   // Unsquelched audio source, toss the first audio buffer after COS goes
   // active to eliminate possible squelch tail on keyup
      D2PRINTF(("I"));
      bTossRxIn = TRUE;
   }
   else if(bTalking && bTossRxIn) {
      bTossRxIn = FALSE;
      D2PRINTF(("i"));
   }
   else if(bTalking && !bMonitor && (!MuteDTMF || DTMF_Digits == 0)) {
      RxAudioReady = -1;   // set the ready bit for all ports
      D2PRINTF(("r"));

   // Count number and type of clients
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(pCC->RxMap & MyBit) {
         // This VoIP client is listening to this port
            if(pCC->CompressionType == RTP_PT_DVI4_8K) {
               AdpcmCount++;
               switch(pCC->Proto) {
                  case PROTO_SF:
                     SfAdpcmCount++;
                     break;

                  case PROTO_RTP:
                     RtpAdpcmCount++;
                     break;

                  default:
                     CORE_DUMP();
                     break;
               }
            }
            else if(pCC->CompressionType == RTP_PT_GSM) {
               GsmCount++;
               switch(pCC->Proto) {
                  case PROTO_SF:
                     SfGsmCount++;
                     break;

                  case PROTO_RTP:
                     RtpGsmCount++;
                     break;

                  case PROTO_ILINK:
                     EchoLinkCount++;
                     break;

                  default:
                     CORE_DUMP();
                     break;
               }
            }
            else {
               LOG_ERROR(("%s: Ignoring %s, unknown compression type %d\n",
                          __FUNCTION__,pCC->Callsign,pCC->CompressionType));
            }
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }

      if(AdpcmCount > 0 || GsmCount > 0) {
#if 0
         LOG_ERROR(("%s: Port %d AdpcmCount: %d, GsmCount: %d\n",
                    __FUNCTION__,PortNum,AdpcmCount,GsmCount));
#endif
      }

      if(AdpcmCount > 0) {
         AdpcmTx(SfAdpcmCount,RtpAdpcmCount);
      }

      if(GsmCount > 0) {
         GsmTx(SfGsmCount,RtpGsmCount,EchoLinkCount);
      }
   }
}

void CPort::GetAudioIn(ClientInfo *p)
{
   int Read;
   int FullBufLen;
   char *pWriteData;
   int i;

#ifndef  DEBUG_PCM_IN
   for( ; ; ) {
      if(AudioIn8Bit) {
         pWriteData = ((char *) AudioInBuf8) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE;
      }
      else {
         pWriteData = ((char *) AudioInBuf) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE * sizeof(int16);
      }
      Read = read(p->Socket,pWriteData,FullBufLen-AudioInAvailable);

      if(Read <= 0) {
         if(errno != EAGAIN) {
            LOG_ERROR(("GetAudioIn(): read() failed: %s",Err2String(errno)));
         }
         p->bReadReady = FALSE;
         break;
      }
      else {
         AudioInAvailable += Read;
         if(AudioInAvailable == FullBufLen) {
            if(AudioIn8Bit) {
            // Convert to 16 bit audio in AudioInBuf
               for(i = 0; i < AUDIO_BUF_SIZE; i++) {
                  AudioInBuf[i] = (int16) ((AudioInBuf8[i] - 128) << 8);
               }
            }
#if 0
            if(PortNum == 0 && fp != NULL) {
               fwrite(AudioInBuf,sizeof(AudioInBuf),1,fp);
            }
#endif
            ProcessAudioIn(p);
            AudioInAvailable = 0;
         }
      }
   }
#endif
}

void GetAudioIn(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;

   pClass->GetAudioIn(p);
}

void CPort::PollCOS(ClientInfo *p)
{
   int CosBit = 0;
   int bCosActive = FALSE;
   PP_DATA_TYPE IRLP_Bits = 0;

   switch(RxCosMethod) {
      case 2:  // IRLP interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_STATUS,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_STATUS) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else {
            bCosActive = (IRLP_Bits & COS_Bit) ? TRUE : FALSE;
         }
         ReleaseIRLP();
         break;

      case 3:  // serial port CTS
         CosBit = TIOCM_CTS;
         break;

      case 4:  // serial port DSR
         CosBit = TIOCM_DSR;
         break;

      case 5:  // serial port DCD
         CosBit = TIOCM_CD;
         break;

      case 6:  // /dev/input/event*
         bCosActive = bDevInCOS;
         break;

      case 7:  // USB device GPIO
#ifdef USB_SUPPORT
         bCosActive = UsbPollCOS();
#endif
         break;
   }

   if(CosBit != 0) {
      if(ioctl(SerialFd,TIOCMGET,&ModemCtrl) == -1) {
         LOG_ERROR(("PollCOS(): ioctl(TIOCMGET) failed: %s",Err2String(errno)));
      }
      else {
         if(ModemCtrl & CosBit) {
            bCosActive = FALSE;
         }
         else {
            bCosActive = TRUE;
         }
      }
   }

   if(InvertCOS) {
      bCosActive = !bCosActive;
   }

   if(g_bCosActive != bCosActive) {
   // COS has changed state
      g_bCosActive = bCosActive;
      DPRINTF(("%s: Port %s COS %sactive\n",__FUNCTION__,PortName,
               bCosActive ? "" : "in"));

      if(bCosActive) {
         CosMap |= MyBit;
         CtoneTimer.tv_sec = 0;
      }
      else {
         CosMap &= ~MyBit;
         RxAudioReady = 0;
         SendCtone();
      }

      if(!DecodeCTCSS) {
         StartTalking(bCosActive);
      }
      else {
      // CTCSS decode enabled
         StartTalking(g_bRxCTCSS && bCosActive);
      }
   }
}

// A callsign is entered as using two digits for each letter and number 
// in the callsign. The first digit is the key on which the letter appears 
// (using 1 for Q and Z), and the second digit is 1, 2, or 3, to indicate 
// which letter is being entered.  Numbers are enter as the digit followed 
// by a 0.  The pound key is used to indicate the end of the string.
// For example, the letter "K" is entered as "52", the letter "Q" is 
// entered as "11", and the digit "7" is entered as "70".
//
// Callsigns need not be entered in full.  If a partial callsign is entered, 
// thebridge will find the first match among the stations currently logged on.
// If no match is found among the stations currently logged on, 
// thebridge will say "NOT FOUND".

char DTMF2CharLookup[] = {
   0x21, // 'A'
   0x22, // 'B'
   0x23, // 'C'
   
   0x31, // 'D'
   0x32, // 'E'
   0x33, // 'F'
   
   0x41, // 'G'
   0x42, // 'H'
   0x43, // 'I'

   0x51, // 'J'
   0x52, // 'K'
   0x53, // 'L'

   0x61, // 'M'
   0x62, // 'N'
   0x63, // 'O'

   0x71, // 'P'
   0x11, // 'Q'
   0x72, // 'R'
   0x73, // 'S'

   0x81, // 'T'
   0x82, // 'U'
   0x83, // 'V'

   0x91, // 'W'
   0x92, // 'X'
   0x93, // 'Y'
   0x12, // 'Z'
};

char *DTMF2Call(char *Digit)
{
   static char Callsign[MAX_CALL_LEN+1];
   int   i;
   unsigned int   j;
   char  FirstDigit;
   char  c = 0;
   char  lookup;

   Callsign[0] = 0;

   for(i = 0; i < MAX_CALL_LEN; ) {
   // Get first digit of pair
      c = *Digit++;
      if(!isdigit(c)) {
      // we're done
         break;
      }
      FirstDigit = c;

   // Get second digit of pair
      c = *Digit++;
      if(!isdigit(c)) {
      // we're done
         break;
      }
      
      if(c == '0') {
      // digit
         Callsign[i] = FirstDigit;
      }
      else {
         lookup = ((FirstDigit << 4) & 0xf0) | (c & 0xf);
         for(j = 0; j < sizeof(DTMF2CharLookup); j++) {
            if(DTMF2CharLookup[j] == lookup) {
               Callsign[i] = (char) j + 'A';
               break;
            }
         }

         if(j == sizeof(DTMF2CharLookup)) {
         // invalid combination
            Callsign[0] = 0;
            break;
         }
      }
   }

   if(c != 0 && c != '#') {
   // invalid combination
      Callsign[0] = 0;
   }
   else {
   // Add terminating null
      Callsign[i] = 0;
   }

   return Callsign;
}


void CPort::SaveDTMF(char Digit)
{
   LastDTMFDigit = TimeNow;
   if(Digit != ' ' && DTMF_Digits < MAX_DTMF_DIGITS) {
      DTMF_Buf[DTMF_Digits++] = Digit;
      DTMF_Buf[DTMF_Digits] = 0;
      D2PRINTF(("Port %s TouchTone decode: %c\n",PortName,Digit));
      DFLUSH();
   }
   EventHook("dtmfdecode %c",Digit == ' ' ? 'N' : Digit);
}


void CPort::PollDTMF(ClientInfo *p)
{
   int wrote;
   int BytesRead;
   static char LastDTMF;
   char CurrentDTMF;
   PP_DATA_TYPE IRLP_Bits = 0;

   switch(DtmfMethod) {
      case 1:  // WB2REM / VA3TO interface
         while((BytesRead = read(SerialFd,&CurrentDTMF,1)) == 1) {
            if((CurrentDTMF >= '0' && CurrentDTMF <= '9') ||
               (CurrentDTMF == '*' || CurrentDTMF == '#') || 
               (CurrentDTMF >= 'A' && CurrentDTMF <= 'D') ||
               (CurrentDTMF == ' '))
            {
               if(LastDTMF != CurrentDTMF) {
                  LastDTMF = CurrentDTMF;
                  SaveDTMF(CurrentDTMF);
               }
            }
         }

         if(BytesRead < 0 && errno != EAGAIN) {
            LOG_ERROR(("PollDTMF(): read failed: %s",Err2String(errno)));
         }

         if(bTxIsKeyed != bTxKeyed) {
            bTxIsKeyed = bTxKeyed;
            wrote = write(SerialFd,bTxKeyed ? "T" : "R",1);
         }
         else {
         // don't need to key or unkey, just poll for DTMF
            wrote = write(SerialFd,"C",1);
         }

         if(wrote != 1) {
            LOG_ERROR(("PollDTMF(): write failed: %s",Err2String(errno)));
         }
         break;

      case 2:  // IRLP interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_STATUS,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_STATUS) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else {
            CurrentDTMF = (IRLP_Bits >> IRLP_DTMF_SHIFT) & 0xf;
            if(CurrentDTMF == 0) {
            // IRLP hardware doesn't include the data valid flag from
            // the DTMF decoder, only the 4 data bits.  Since the decoder 
            // used outputs a "D" (0x0) when no digit is decoded it must
            // be tossed since we can't tell if it's valid or not.
               CurrentDTMF = ' ';
            }
            else if(CurrentDTMF < 10) {
               CurrentDTMF += '0';
            }
            else if(CurrentDTMF == 10) {
               CurrentDTMF = '0';
            }
            else if(CurrentDTMF == 11) {
               CurrentDTMF = '*';
            }
            else if(CurrentDTMF == 12) {
               CurrentDTMF = '#';
            }
            else {
               CurrentDTMF = 'A' + (CurrentDTMF - 13);
            }

            if(LastDTMF != CurrentDTMF) {
               LastDTMF = CurrentDTMF;
               SaveDTMF(CurrentDTMF);
            }
         }
         ReleaseIRLP();
         break;
   }
}


int CPort::Audio_Handler(ClientInfo *p)
{
   audio_buf_info BufInfo;
   int bWroteOutput = FALSE;

   bAudioOutCalled = FALSE;

   if(p->bReadReady) {
      GetAudioIn(p);
   }

   if(LastID != 0 && (TimeNow.tv_sec - LastID) > IdInterval) {
   // time to ID
      ID();
   }

   if(bTxKeyed && !bTimeOut && TxTimeout != 0 &&  
      (TimeNow.tv_sec - TxKeyTime) > TxTimeout) 
   do {  // Timeout !
      LOG_ERROR(("%s: Port %s Tx timeout\n",__FUNCTION__,PortName));
      if(TimeoutTone != 0) {
         char *ToneSpec = FindToneSpec(TimeoutTone);

         if(ToneSpec == NULL) {
            LOG_ERROR(("%s: Timeout tone ID #%d is not defined, disabling.\n",
                       __FUNCTION__,TimeoutTone));
            TimeoutTone = 0;
            break;
         }

         if((TimeoutToneGen = new CToneGen()) == NULL) {
            LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
            break;
         }

         if(TimeoutToneGen->ToneGenSetup(ToneSpec,NULL) != 0) {
            LOG_ERROR(("%s#%d: ToneGenSetup failed\n",__FUNCTION__,__LINE__));
            delete TimeoutToneGen;
            TimeoutToneGen = NULL;
         }
      }
      bTimeOut = TRUE;
   // suppress courtesy tone when we resume from a timeout
      VoipCtoneTimer.tv_sec = 0;
   } while(FALSE);

   if(p->bWriteReady) {
   // Audio output needs another buffer
      WriteWakeups++;
      D2PRINTF(("R"));
      bWroteOutput = WriteAudioOut(p);
   }


   PollCOS(p);

   if(DtmfMethod == 2 ||
      (DtmfMethod == 1 && TimeLapse(&LastDTMFPoll) >= DTMFPollTime))
   {
      PollDTMF(p);
      LastDTMFPoll = TimeNow;
   }

   if(CtoneTimer.tv_sec != 0) {
      if(TimeLapse(&CtoneTimer) >= CtoneDelay) {
         SendCtone();
         CtoneTimer.tv_sec = 0;
      }
   }

   if(VoipCtoneTimer.tv_sec != 0) {
      if(TimeLapse(&VoipCtoneTimer) >= VoipCtoneDelay) {
         VoipCtoneTimer.tv_sec = 0;
         D2PRINTF(("%s: calling GenVoipCtone\n",__FUNCTION__));
         GenVoipCtone();
      }
   }

   if(!bAudioOutCalled && pPlayHead != NULL &&
      (!(pPlayHead->Flags & PLAY_FLAG_TTS_OUT) || !bTTS_Running))
   {  // Output audio wasn't running and there's an announcement to play
      // and it's either not an TTS rendered announcement or the TTS
      // system has finished rendering it...  start the output audio stream
      bWroteOutput = WriteAudioOut(pAudioC);
   }

   if(!bTxKeyed && bTimeOut) {
      if((RxMap & CosMap) == 0 && !ActivePorts()) {
         LOG_NORM(("%s: Cleared tx timeout\n",__FUNCTION__));
         bTimeOut = 0;
      }
   }

   if(!bAudioOutCalled && SourcesReady(NULL)) {
      bWroteOutput = WriteAudioOut(pAudioC);
   }

   if(bTxKeyed && !bWroteOutput) {
   // Check to see if the output buffer has drained to zero
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETOSPACE,&BufInfo) == -1) {
         LOG_ERROR(("Audio_Handler(): SNDCTL_DSP_GETOSPACE failed: %s",
                    Err2String(errno)));
      }
      else if(BufInfo.fragments == BufInfo.fragstotal) {
      // The output buffer is empty, unkey the transmitter
         if(HangTime == 0) {
         // Transmitter has no hang time, unkey now
            KeyTx(FALSE);
         }
         else if(HangTimer.tv_sec == 0) {
         // Start HangTimer
            HangTimer = TimeNow;
         }
         else if(TimeLapse(&HangTimer) >= HangTime) {
            HangTimer.tv_sec = 0;
            KeyTx(FALSE);
         }
      }
   }

   DFLUSH();

   return FALSE;
}

int Audio_Handler(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;

   return pClass->Audio_Handler(p);
}

void CPort::AudioCleanup(ClientInfo *p)
{
   KeyTx(FALSE);
   
   if(TxKeyMethod == 1 && SerialFd != -1) {
   // WB2REM / VA3TO interface
      usleep(1000*DTMFPollTime);
      write(SerialFd,"R",1);
   }
}

void AudioCleanup(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;
   pClass->AudioCleanup(p);
}

int CPort::SerialInit()
{
   int Ret = 0;
   struct termios options;
   int i;
   
   if(SerialDevName == NULL) {
      LOG_ERROR(("SerialInit(): Error SerialDevName is not defined.\n"));
      Ret = ERR_SERIAL_DEV_OPEN;
   }
   else if((SerialFd = open(SerialDevName,SERIAL_OPEN_FLAGS)) < 0) {
      LOG_ERROR(("SerialInit(): Couldn't open \"%s\": %s",SerialDevName,
                 Err2String(errno)));
      Ret = ERR_SERIAL_DEV_OPEN;
   }
   else if(tcgetattr(SerialFd,&options) < 0) {
      LOG_ERROR(("SerialInit(): Unable to read port configuration: %s",
                 Err2String(errno)));
      Ret = ERR_SERIAL_GET_ATTR;
   }
   else {
      options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);    /* raw input */
      options.c_iflag &= ~(IXON | IXOFF | IXANY);  /* disable software flow control */
      options.c_iflag |=  IGNBRK;            /* ignore break */
      options.c_iflag &= ~ISTRIP;            /* do not strip high bit */
      options.c_iflag &= ~(INLCR | ICRNL);   /* do not modify CR/NL   */
      options.c_cflag |=  (CLOCAL | CREAD);  /* enable the receiver and set local mode */
      options.c_cflag &= ~CRTSCTS;           /* disable hardware flow control */
      options.c_cflag &= ~CSIZE;             /* Mask the character size bits */
      options.c_cflag |= CS8;                /* 8 data bits */
      options.c_cflag &= ~PARENB;            /* no parity */
      options.c_cflag &= ~CSTOPB;            /* 1 stop bit */

      for(i = 0; BaudRateTbl[i].Baudrate != 0; i++) {
         if(BaudRateTbl[i].Baudrate == SerialBaudRate) {
            if(cfsetspeed(&options,BaudRateTbl[i].Baudrate) == -1) {
               LOG_ERROR(("SerialInit(): cfsetspeed() failed: %s",
                          Err2String(errno)));
               Ret = ERR_SERIAL_SET_BAUD;
            }
            break;
         }
      }

      if(BaudRateTbl[i].Baudrate == 0) {
         LOG_ERROR(("SerialInit(): Baudrate %d not supported.\n"));
         Ret = ERR_SERIAL_SET_BAUD;
      }
   }

   if(Ret == 0 && tcsetattr(SerialFd,TCSANOW,&options) < 0) {
      LOG_ERROR(("SerialInit(): tcsetattr() failed: %s",Err2String(errno)));
      Ret = ERR_SERIAL_SET_ATTR;
   }

   if(Ret != 0) {
      close(SerialFd);
      SerialFd = -1;
   }

   return Ret;
}


int CPort::InitToneBuf(struct config_entry *p, char *Arg)
{
   double ToneFreq;
   double Cycles;
   double BufDuration;
   int i;
   int Ret = 1;   // assume the worse
   void *var_ptr;       // pointer to variable

   if(Arg == NULL || p == NULL) {
      Ret = 0;
   }
   else {
      var_ptr = ConfigGetVarPtr(p);
      if(sscanf(Arg,p->var_type,var_ptr) == 1) {
      // fudge the ID tone frequency so an integer number of cycles fit
      // into our tone buffer
      
         BufDuration = AUDIO_BUF_SIZE/8000.0;
         Cycles = floor(CWToneFreq * BufDuration);
         ToneFreq = Cycles / BufDuration;
      
         for(i = 0; i < AUDIO_BUF_SIZE; i++) {
            ToneBuf[i] = (int16) (CWToneLevel * sin(i*Cycles*2*M_PI / AUDIO_BUF_SIZE));
         }
      
      // Calculate number of cycles of tone in a dit.
      //
      // The word PARIS was chosen as the standard length for CW code speed. 
      // Each dit counts for one count, each dah counts for three counts, 
      // intra-character spacing is one count, inter-character spacing is three 
      // counts and inter-word spacing is seven counts, so the word PARIS is 
      // exactly 50 counts.
      // So 1 wpm = 50 counts/minute or 50/60 counts/second.
      // or Samples/dit = (60/50 secs/dit) * (AUDIO_BUF_SIZE samples) / 
      //                  (BufDuration secs);
      
         DitSamples = (int) ((60.0 * AUDIO_BUF_SIZE) / 
                             (50.0 * BufDuration * CWSpeed));  
      
      // Calculate attach and decay time for tone envelope
      // 10% of dit time ramping up, 80% @ full tone level, 10% ramping down
      
         CwRampCount = DitSamples * 10 / 100;
         Ret = 0;
      }
   }

   return Ret;
}

int InitToneBuf(struct config_entry *p, char *Arg, AfCmd Cmd, int x)
{
   int Ret = 0;

   switch(Cmd) {
      case AF_SET_VAR:
         if(Ports[CurrentPort] != NULL) {
            Ret = Ports[CurrentPort]->InitToneBuf(p,Arg);
         }
         break;

      case AF_SET_DEFAULT:
         break;

      case AF_DISPLAY_VAR:
         break;
   }

   return Ret;
}


void CPort::GenCTCSS(int16 *CTCSSBuf)
{
   int i;
   double CTCSSAngleDelta;

   CTCSSAngleDelta = 2 * M_PI * EchoCTCSSFreq / 8000.0;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      CTCSSBuf[i] = (int16) (CTCSSLevel * sin(CTCSSAngle));
      CTCSSAngle += CTCSSAngleDelta;
   }
}

void CPort::SendCWMsg(char *Msg)
{
   CWChar = Msg;
   if(!pAudioC->bWriteWait) {
      WriteAudioOut(pAudioC);
   }
}

char *CwLookUp(char CwChar)
{
   int   i;
   char *Ret = "";

   if(isalpha(CwChar)) {
      Ret = CwAlphaLookup[toupper(CwChar)-'A'];
   }
   else if(isdigit(CwChar)) {
      Ret = CwNumericLookup[CwChar-'0'];
   }
   else for(i = 0; CwMiscLookup[i].Char != 0; i++) {
      if(CwChar == CwMiscLookup[i].Char) {
         Ret = CwMiscLookup[i].Morse;
         break;
      }
   }

   return Ret;
}

void CPort::GenCW(int16 *CwBuf)
{
   int i = 0;
   int CopyCount;
   char c;

   while(i < AUDIO_BUF_SIZE) {
      if(ElementCount == 0) {
         if(bCwToneOn) {
         // Send intra-character space
            bCwToneOn = FALSE;
            ElementCount = DitSamples;
            bElementSpace = TRUE;
            RampCount = CwRampCount;
         }
         else if(bElementSpace) {
         // next element
            bElementSpace = FALSE;
            if((c = *Element++) != 0) {
               bCwToneOn = TRUE;
               RampCount = CwRampCount;
               if(c == '-') {
                  ElementCount = DitSamples * 3;
               }
               else if(c == '.') {
                  ElementCount = DitSamples;
               }
            }
            else {
            // Next character, already did a one dit time for the element
            // space, do 2 more for character space
               ElementCount = DitSamples * 2;
            }
         }
         else {
         // character space complete, do next character
            c = *CWChar;

            if(c == 0) {
            // That's it
               CWChar = NULL;
               if(i > 0) {
                  memset(&CwBuf[i],0,(AUDIO_BUF_SIZE-i) * sizeof(int16));
                  i = AUDIO_BUF_SIZE;
               }
               break;
            }
            else if(c == ' ') {
            // Word space, 3 dit time space already sent.  Send 4 more
               ElementCount = DitSamples * 4;
               CWChar++;
            }
            else {
               for( ; ; ) {
                  if((c = *CWChar) == 0) {
                  // end of string
                     CWChar = NULL;
                     break;
                  }
                  CWChar++;
                  Element = CwLookUp(c);
                  if(*Element == '-' || *Element == '.') {
                  // Start sending the next character
                     bElementSpace = TRUE;
                     break;
                  }
               }
            }
         }
      }

      if(RampCount > 0) {
      // Ramping tone level up or down
         if(bCwToneOn) {
         // Ramping up
            while(ElementCount-- > 0 && i < AUDIO_BUF_SIZE && RampCount > 0) {
               CwBuf[i] = (int16) ((ToneBuf[i] * (CwRampCount - RampCount)) / 
                                   CwRampCount);
               i++;
               RampCount--;
            }
         }
         else {
            // Ramping tone level down
            while(ElementCount-- > 0 && i < AUDIO_BUF_SIZE && RampCount > 0) {
               CwBuf[i] = (int16) ((ToneBuf[i] * RampCount--) / CwRampCount);
               i++;
            }
         }
      }
      else if(bCwToneOn) {
      // Copy full level tone
         CopyCount = min(ElementCount,AUDIO_BUF_SIZE-i);
         memcpy(&CwBuf[i],&ToneBuf[i],CopyCount*sizeof(int16));
         i += CopyCount;
         ElementCount -= CopyCount;
      }
      else {
      // No tone, just zero the buffer
         CopyCount = min(ElementCount,AUDIO_BUF_SIZE-i);
         memset(&CwBuf[i],0,CopyCount*sizeof(int16));
         i += CopyCount;
         ElementCount -= CopyCount;
      }
   }
}


/* tone generation

   .tonegen [<tone1 freq>] [<tone2 freq>] [<level>] <duration> ...

   # <id>         Play saved ToneSpec #<id>  
   TF1 <freq>     Set Tone 1 frequency, 0 = no tone
   EF1 <freq>     Set Tone 1 end frequency for sweeps (starts at TF1)
   TL <level>     Set Tone 1 and 2 level

   TF2 <freq>     Set Tone 2 frequency, 0 = no tone
   TL2 <level>    Set Tone 2 level
   DUR <duration> duration of tone segment in milliseconds.
   
   RPT <count>    repeat pattern from the beginning or last rpt count times,
                  -1 means repeat until cancelled.

   Each tone segment must have an duration, all other fields are optional.
   
   Tone frequencies are cleared set to zero after each segment is generated.
   Tone levels are only changed if explicitly modified.
   
   dial tone is a continuous tone having frequencies of 350 and 440 Hz at a 
   level of -13 dBm

   ringback tone is defined as comprising frequencies of 440 and 480 Hz at a 
   level of -19 dBm and a cadence of 2 seconds ON and 4 seconds OFF

   busy tone is defined as having frequency components of 480 and 620 Hz at a 
   level of -24 dBm and a cadence of half a second ON and half a second OFF

   reorder tone, also called "fast busy" tone, contains the same frequency 
   components as busy tone at a similar level but with a cadence of 0.25 of 
   a second ON and 0.25 of a second OFF
*/
void CmdToneGen(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   char *pErrPos;
   char *ToneSpec = Arg;
   int ToneID;

   do {
      if(*Arg == '#') {
      // play a stored tonespec
         if(sscanf(&Arg[1],"%d",&ToneID) != 1 || ToneID <= 0) {
            BufPrintf(p,"Error: invalid tone ID \"%s\"\r",Arg);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }

         if((ToneSpec = FindToneSpec(ToneID)) == NULL) {
            BufPrintf(p,"Error: tone #%d is not defined.\r",ToneID);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }
      }
      CToneGen *pGenerator = new CToneGen();

      if(pGenerator == NULL) {
         LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
         BufPrintf(p,"Error: new failed\r");
         break;
      }

      if(pGenerator->ToneGenSetup(ToneSpec,&pErrPos)) {
      // error
         BufPrintf(p,"Error: couldn't parse \"%s\"\r",pErrPos);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      CPort *pPort = GetPortPtr(CurrentPort);
      pPort->AddToneGen(pGenerator);
   } while(FALSE);
}

struct DtmfCode {
   int Row;
   int Col;
} DtmfFreqLookup[] = {
   { 941, 1336},  // 0
   { 697, 1209},  // 1
   { 697, 1336},  // 2
   { 697, 1477},  // 3
   { 770, 1209},  // 4
   { 770, 1336},  // 5
   { 770, 1477},  // 6
   { 852, 1209},  // 7
   { 852, 1336},  // 8
   { 852, 1477},  // 9
   { 941, 1209},  // *
   { 941, 1477},  // #
   { 697, 1633},  // A
   { 770, 1633},  // B
   { 852, 1633},  // C
   { 941, 1633}   // D
};

// .dtmfgen <dtmf digits>
// One or more '+' maybe used to extend the duration and pause for the next 
// digit.  For example +++*1234 = generates a '*' digit with 4 times the 
// normal duration and 4 times the normal pause before generating 1234 with
// the normal timings
void CmdDtmfGen(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   char ToneSpec[1024];
   int ToneSpecLen = 0;
   int Index = -1;
   char Digit;
   int Count;
   int TimeMultiplier = 1;
   
   while(*Arg != 0) {
      switch((Digit = tolower(*Arg++))) {
         case ' ':
            continue;

         case 'a':
         case 'b':
         case 'c':
         case 'd':
            Index = 12 + Digit-'a';
            break;

         case '0':
         case '1':
         case '2':
         case '3':
         case '4':
         case '5':
         case '6':
         case '7':
         case '8':
         case '9':
            Index = Digit-'0';
            break;

         case '*':
            Index = 10;
            break;

         case '#':
            Index = 11;
            break;

         case '+':
            TimeMultiplier++;
            continue;
            break;
            
         default:
            Index = -1;
            break;
      }

      if(Index == -1) {
         BufPrintf(p,"Error: invalid DTMF digit \"%c\"\r",Digit);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      Count = snprintf(&ToneSpec[ToneSpecLen],sizeof(ToneSpec)-ToneSpecLen,
                       "TF1 %d TF2 %d dur %d dur %d ",
                       DtmfFreqLookup[Index].Row,DtmfFreqLookup[Index].Col,
                       DtmfEncodeDuration*TimeMultiplier,
                       DtmfEncodePause*TimeMultiplier);

      if(Count == -1 || Count >= (int) sizeof(ToneSpec)-ToneSpecLen) {
      // Pre-C99 style snprintf returns -1 on buffer overflow.  
      // C99 style snprintf returns the number of characters that would
      // have been written on buffer overflow.
      // In either case the output was truncated and p->Count is bogus.
         BufPrintf(p,"Error: DTMF buffer overflow\r");
         Index = -1;
         break;
      }
      else {
         ToneSpecLen += Count;
      }
      TimeMultiplier = 1;
   }

   if(Index != -1) {
      LOG_ERROR(("%s: spec \"%s\"\n",__FUNCTION__,ToneSpec));
      char *pErrPos;
      CToneGen *pGenerator = new CToneGen();
      if(pGenerator->ToneGenSetup(ToneSpec,&pErrPos)) {
      // error
         BufPrintf(p,"Error: couldn't parse \"%s\"\r",pErrPos);
         Rcode = TBD_ERR_INVALID_ARG;
      }
      else {
         CPort *pPort = GetPortPtr(CurrentPort);
         pPort->AddToneGen(pGenerator);
      }
   }
}

static void digit_delivery(void *data, const char *digits, int len)
{
    int i;
    CPort *pClass = (CPort *) data;

    for(i = 0; i < len; i++) {
       if(digits[i] != 0) {
          pClass->SaveDTMF(digits[i]);
       }
    }
}

int CPort::EndPointInit()
{
   int Ret = 0;
   int SampleSize = 16;
   int SampleRate = PCMRate;
   int Stereo = 0;
   int Fmt = AFMT_S16_LE;
   int bits;
   audio_buf_info BufInfo;
#ifdef SNDCTL_DSP_SETBLKSIZE
   int BlkSz = 1280;
#else
   long frag_size = 0x50008;  // 5 x 256 byte fragments
#endif

   LOG_NORM(("%s: called, TxKeyMethod: %d, RxCosMethod: %d\n",__FUNCTION__,
             TxKeyMethod,RxCosMethod));

   bEchoIrlpMode = TRUE;

   if(DecodeCTCSS) {
      CtcssInit(EchoCTCSSFreq);
   }

   for( ; ; ) {
      if(PortName == NULL) {
         char Temp[32];
      // No port name was specified, use a numeric number
         sprintf(Temp,"%d",PortNum);
         PortName = strdup(Temp);
      }
      if(AudioIn8Bit) {
         if((AudioInBuf8 = (char *) malloc(AUDIO_BUF_SIZE)) == NULL) {
            LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            Ret = ERR_MALLOC;
            break;
         }
      }

      if((AudioInBuf = (int16 *) malloc(AUDIO_BUF_SIZE*2)) == NULL) {
         LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
         Ret = ERR_MALLOC;
         break;
      }

      if(TxKeyMethod == 5) {
#ifdef USB_SUPPORT
         if((Ret = UsbInit()) != 0) {
            break;
         }
#else
         LOG_ERROR(("%s: Error - thelinkbox was not configured with USB support.\n",
                    __FUNCTION__));
         Ret = ERR_USB_DEV_OPEN;
         break;
#endif
      }

      if(RxCosMethod == 6) {
#ifdef __linux__
         if((Ret = InitDeviceInput()) != 0) {
            LOG_ERROR(("%s: Couldn't find /dev/inputX entry for RxCosMethod 6\n",
                       __FUNCTION__));
            break;
         }
#else
         LOG_ERROR(("Error: RxCosMethod is only supported on Linux\n"));
         Ret = ERR_INPUT_DEV_OPEN;
         break;
#endif
      }

      pAudioC = CreateNewClient();

      if(pAudioC == NULL) {
         Ret = ERR_MALLOC;
         break;
      }

      pAudioC->Cleanup = ::AudioCleanup;
      pAudioC->State = ::Audio_Handler;
      pAudioC->p = this;

      if((pAudioC->Socket = open(AudioDevice,AUDIO_OPEN_FLAGS)) < 0) {
         LOG_ERROR(("%s#%d: open(\"%s\") failed: %s",__FUNCTION__,__LINE__,AudioDevice,
                   Err2String(errno)));
         Ret = ERR_AUDIO_DEV_OPEN;
         break;
      }

      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETDUPLEX,NULL) == -1) {
      // Failure is not necessarily fatal.
      // FreeBSD doesn't require this call for fullduplex
         LOG_WARN(("EndPointInit(): SNDCTL_DSP_SETDUPLEX failed: %s",
                   Err2String(errno)));
      }

      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETFMTS,&bits) == -1) {
         LOG_WARN(("EndPointInit(): SNDCTL_DSP_GETFMTS failed: %s",
                   Err2String(errno)));
      }
      else if(!(bits & AFMT_S16_LE)) {
         LOG_ERROR(("EndPointInit(): Error AFMT_S16_LE not supported\n"));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SAMPLESIZE,&SampleSize) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SAMPLESIZE failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_STEREO,&Stereo) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_STEREO failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SPEED,&SampleRate) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SPEED failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      else if(SampleRate != PCMRate) {
         LOG_WARN(("%s: Asked for a sampling rate of %d, got %d\n",
                   __FUNCTION__,PCMRate,SampleRate));
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETFMT,&Fmt) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETFMT failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }

#ifdef SNDCTL_DSP_SETBLKSIZE
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETBLKSIZE,&BlkSz) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETBLKSIZE failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
#else
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETFRAGMENT,&frag_size) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETFRAGMENT failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_SET_FRAG;
         break;
      }
#endif
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETOSPACE,&BufInfo) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_GETOSPACE failed: %s",
                    Err2String(errno)));
      }
      else {
         LOG_ERROR(("Fragsize: %d, fragstotal: %d, bytes: %d\n",
                    BufInfo.fragsize,BufInfo.fragstotal,BufInfo.bytes));
      }
      
      if((hGSM_Encode = gsm_create()) == NULL) {
         Ret = ERR_GSM_INIT;
         break;
      }

      if((TxKeyMethod == 1 || TxKeyMethod == 3 || TxKeyMethod == 4 ||
          RxCosMethod == 3 || RxCosMethod == 4 || RxCosMethod == 5))
      {  // serial port needed by TxKeyMethod or RxCosMethod

         if(TxKeyMethod == 3) {
            PTT_Bit = TIOCM_RTS;
         }
         else if(TxKeyMethod == 4) {
            PTT_Bit = TIOCM_DTR;
         }

         if(!FindSerialMaster()) {
         // This serial port has not been opened yet
            if((Ret = SerialInit()) != 0) {
               break;
            }
         }
      }

      if(TxKeyMethod == 2 || RxCosMethod == 2) {
      // IRLP port needed by TxKeyMethod or RxCosMethod
         if(PTT_Bit == -1) {
            PTT_Bit = IRLP_PTT;
         }

         if(COS_Bit == -1) {
            COS_Bit = IRLP_COS;

         }
         if((IRLPfd = open(IRLP_DEVICE,O_RDWR)) < 0 ) { 
            LOG_ERROR(("EndPointInit(): open failed for \"%s\": %s",
                       IRLP_DEVICE,Err2String(errno)));
            Ret = ERR_IRLP_DEV_OPEN;
            break;
         }

         if(ShareIRLP_Port) {
            if((LockIRLPfd = open(IRLP_LOCKFILE,O_WRONLY|O_CREAT, 0664)) < 0 ) { 
               LOG_ERROR(("EndPointInit(): open failed for \"%s\": %s",
                          IRLP_LOCKFILE,Err2String(errno)));
               Ret = ERR_IRLP_LOCKFILE_OPEN;
               break;
            }
         }
         else {
#ifdef __linux__
            if(ioctl(IRLPfd,PPCLAIM) != 0) {
                LOG_ERROR(("EndPointInit(): ioctl(PPCLAIM) failed: %s",
                           Err2String(errno)));
                Ret = ERR_IRLP_DEV_CLAIM;
                break;
            } 
#endif
         }

         if(InitPPData != -1) {
            char  IRLP_Bits = (char) InitPPData;

            if(ClaimIRLP()) {
               LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
            }
            else {
               if(ioctl(IRLPfd,PP_WR_DATA,&IRLP_Bits)) {
                  LOG_ERROR(("%s: ioctl(PPWDATA) failed: %s",__FUNCTION__,
                             Err2String(errno)));
               }
               ReleaseIRLP();
            } 
         }

         LOG_NORM(("%s: IRLPfd: %d\n",__FUNCTION__,IRLPfd));
      }

      if(DtmfMethod == 3) {
         dtmf_rx_init(&DtmfRxState,digit_delivery,this);
         dtmf_rx_parms(&DtmfRxState,0,Twist,ReverseTwist);
      }

      avl_insert(ClientTree,pAudioC);
      pAudioC->bInClientTree = TRUE;
      pAudioC->bReadWait = TRUE;
      GetAudioIn(pAudioC);
      break;
   }

   if(Ret != 0) {
   // Init error, clean up
      if(pAudioC != NULL) {
         if(pAudioC->Socket >= 0) {
            close(pAudioC->Socket);
         }
         DeleteClient(pAudioC);
         pAudioC = NULL;
      }

      if(IRLPfd != -1) {
         close(IRLPfd);
         IRLPfd = -1;
      }

      if(LockIRLPfd != -1) {
         close(LockIRLPfd);
         LockIRLPfd = -1;
      }
   }
   else {
      KeyTx(FALSE);
   }


   LOG_NORM(("%s: returning %d\n",__FUNCTION__,Ret));

   return Ret;
}

int EndPointInit() 
{
   int Ret = -1;
   int i;

#if 0
   fp = fopen("test.raw","w");
#endif

// !!!
   CurrentPort = 0;

   for(i = 0; i < NumPorts; i++) {
      LOG_ERROR(("%s: Initializing %s port (%d)\n",__FUNCTION__,
					  Ports[i]->PortName,i));
      if((Ret = Ports[i]->EndPointInit()) != 0) {
         break;
      }
   }

   return Ret;
}

void ExecuteCommand(char *Command)
{
   ConfClient cc;

   memset(&cc,sizeof(cc),0);
   cc.bCmdLine = TRUE;
   cc.bSysop = TRUE;
   cc.bAdmin = TRUE;
   cc.bRfUser = TRUE;
   cc.Callsign = "RF";
   cc.CallPlus = cc.Callsign;
   cc.pCS = piLinkConf;

   LOG_NORM(("RF user executing command \"%s\"\n",Command));
   ConferenceCmd(cc.pCS->pAudio,&cc,Command);
}


char *StdTbdCommandText[] = {
   // command not found or ambiguous
   "unknown command",      
   
   // command requires disk and disk is disabled
   "no disk is available"  
   
   // Specified station not found
   "station not found",    
   
   // no info available for station (.info cmd)
   "no information on station is available", 

   // invalid command argument count
   "error, invalid number command arguments",

   // Already connected (.connect cmd)
   "you are already connected to that station",

   // no one is talking (.mute cmd)
   "no one is talking",

   // invalid command argument
   "command argument is not valid",

   // error opening a file
   "error can not open file",

   // data is chat text
   "",

   // timeout waiting for response from tbd ... it's probably not running
   "error timeout",

   // last "command" was sent as chat text
   ""
};

int SayConnectResult(ClientInfo *p,ConfClient *pCC,char *Arg)
{
   int Ret = FALSE;
   CPort *pPort = GetPortPtr(CurrentPort);

   if(LastRcode == TBD_OK) {
      if(StationDisconnected != NULL) {
         pPort->SayPhrase("connecting to ");
         Ret = TRUE;
      }
   }

   return Ret;
}

int SayDisconnectResult(ClientInfo *p,ConfClient *pCC,char *Arg)
{
   int Ret = FALSE;
   CPort *pPort = GetPortPtr(CurrentPort);
   
   if(LastRcode == TBD_OK) {
      if(StationDisconnected != NULL) {
         pPort->SayCall(StationDisconnected);
         pPort->SayPhrase("disconnected");
         Ret = TRUE;
      }
   }

   return Ret;
}


struct SayResultHandler {
   const char *Cmd;
   int (*Handler)(ClientInfo *p,ConfClient *pCC,char *Arg);
} SayResultHandlers[] = {
   {"connect", SayConnectResult},
   {"disconnect", SayDisconnectResult},
   { NULL}  // end of table
};

int CPort::Shutup()
{
   CPort *pPort = GetPortPtr(CurrentPort);
   int Ret = FALSE;

   if(pPlayHead == NULL) {
      Ret = TRUE;
      while(pPort->pPlayHead != NULL) {
         pPort->FreeAnnoucement();
      }
   }

   return Ret;
}

void CmdShutup(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CPort *pPort = GetPortPtr(CurrentPort);

   if(pPort->Shutup()) {
      BufPrintf(p,"ok\r");
   }
   else {
      BufPrintf(p,"no announcements are playing\r");
   }
}

void CmdSayResult(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   char ErrMsg[120];
   int  bHandled = FALSE;
   int i;
   CPort *pPort = GetPortPtr(CurrentPort);

// Search for a custom say result handler for this command

   for(i = 0; SayResultHandlers[i].Cmd != NULL; i++) {
      if(strcmp(SayResultHandlers[i].Cmd,LastCmd) == 0) {
      // Found it
         bHandled = SayResultHandlers[i].Handler(p,pCC1,Arg);
         break;
      }
   }

   if(!bHandled) {
      if(TTS_ExePath) {
      // We have TTS

         if(LastCmdText != NULL) {
         // Something to say, just say it
            pPort->SayPhrase(LastCmdText);
         }
         else if(LastRcode == TBD_OK) {
            pPort->SayPhrase("O K");
         }
         else if(LastRcode > TBD_ERR_BASE && LastRcode < TBD_LAST_ERR_CODE) {
            pPort->SayPhrase(StdTbdCommandText[LastRcode - TBD_ERR_BASE - 1]);
         }
         else {
            snprintf(ErrMsg,sizeof(ErrMsg),"command failed %s",
                     Err2String(LastRcode));
            pPort->SayPhrase(ErrMsg);
         }
      }
      else {
      // No TTS
         if(LastRcode == TBD_OK) {
            pPort->SayPhrase("O");
            pPort->SayPhrase("K");
         }
         else {

         }
      }
   }
}

int Comment(char *cp)
{
   while(*cp && isspace(*cp))
      cp++;
   if(*cp == ';' || *cp == 0 )
      return 1;
   return 0;
}

#define MAX_LINE_LEN 120
int CPort::LoadCommandFile(struct config_entry *p, char *Arg)
{
   int Ret = 0;
   int LineNum = 0;
   FILE *fp = NULL;
   char  Line[MAX_LINE_LEN];
   char  DTMF_String[MAX_LINE_LEN+1];
   char  CommandString[MAX_LINE_LEN+1];
   char  *cp;
   char  *cp1;
   DTMF_Command *pDTMFCommand;
   DTMF_Command *pTail = (DTMF_Command *) &pDTMFCommands;
   char Filename[PATH_MAX];
   
   if(Arg != NULL) {
      while((pDTMFCommand = pDTMFCommands) != NULL) {
      // Clean up old commands before loading new ones
         pDTMFCommands = pDTMFCommand->Link;
         if(pDTMFCommand->DTMF_String != NULL) {
            free(pDTMFCommand->DTMF_String);
         }
         if(pDTMFCommand->Command != NULL) {
            free(pDTMFCommand->Command);
         }
         free(pDTMFCommand);
      }

      if(Arg[0] == '/') {
         strcpy(Filename,Arg);
      }
      else {
         snprintf(Filename,PATH_MAX,"%s/%s",WorkingDir,Arg);
      }

      if((fp = fopen(Filename,"r")) == NULL) {
         LOG_ERROR(("%s: Unable to open \"%s\" - %s\n",__FUNCTION__,Filename,
                    Err2String(errno)));
      }
   }
   
   while(fp != NULL) {
      if(fgets(Line,sizeof(Line),fp) == NULL) {
         if(ferror(fp)) {
            LOG_ERROR(("%s: read error while reading \"%s\" - %s\n",
                       __FUNCTION__,Filename,Err2String(errno)));
            Ret = errno;
         }
         break;
      }
      LineNum++;
      if((cp = strrchr(Line,'\n')) != NULL) {
         // remove terminating <lf>
         *cp = 0;
      }

      if((cp = strrchr(Line,'\r')) != NULL) {
      // remove terminating <cr> just in case some one uses an editor that
      // changes the end of line convention in *nix
         *cp = 0;
      }

      if(Comment(Line)) {
      // ignore this line if it's a comment
         continue;
      }

      if(sscanf(Line," %[^= ] = %[^\n]",DTMF_String,CommandString) != 2) {
      // line didn't convert
         LOG_WARN(("nError: Unable to parse line:\n  %s\n",Line));
         continue;
      }

   // Remove spaces from DTMF string and test for valid characters
      cp1 = cp = DTMF_String;
      while(*cp) {
         *cp = toupper(*cp);
         if((*cp >= 'A' && *cp <= 'D') || (*cp >= '0' && *cp <= '9') ||
            *cp == '#' || *cp == '*' || *cp == '?' || *cp == '+')
         {  // Valid DTMF or wildcard character, copy it

            *cp1++ = *cp++;
         }
         else if(*cp == ' ' || *cp == '\t') {
         // space or tab, ignore it
            cp++;
         }
         else {
         // error
            LOG_WARN(("Error: \"%c\" is not an DTMF character, ignoring line:\n"
                      "%s\n",*cp,Line));
         }
      }

      if(*cp) {
         continue;
      }

      *cp1 = 0;

      cp = CommandString;
   // Skip white space at beginning of the command string
      while(isspace(*cp)) {
         cp++;
      }

   // Good command, add it to the command chain

      if((pDTMFCommand = (DTMF_Command *) malloc(sizeof(DTMF_Command))) == NULL) {
         LOG_ERROR(("%s:#d: malloc failed.\n",__FUNCTION__,__LINE__));
         break;
      }

      pDTMFCommand->Link = NULL;
      pDTMFCommand->DTMF_String = strdup(DTMF_String);
      pDTMFCommand->Command = strdup(cp);
      pTail->Link = pDTMFCommand;
      pTail = pDTMFCommand;
   }

   if(fp != NULL) {
      fclose(fp);
   }

   return Ret;
}

int LoadCommandFile(struct config_entry *p, char *Arg, AfCmd Cmd, int x)
{
   CPort *pPort;
   void *pVar;
   int Ret = -1;

   switch(Cmd) {
      case AF_SET_VAR:
         if(ConfigPass > 0) {
         // NB: we need to call ConfigGetVarPtr even though we don't use 
         // the pointer to make sure the class gets created in case 
         // DTMFCommandFile is the first port configuration variable in the 
         // config file.
            pVar = ConfigGetVarPtr(p);
            pPort = GetPortPtr(CurrentPort);
            Ret = pPort->LoadCommandFile(p,Arg);
         }
         break;

      case AF_SET_DEFAULT:
         break;

      case AF_DISPLAY_VAR:
         break;
   }
   
   return Ret;
}


void CmdSay(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CPort *pPort = GetPortPtr(CurrentPort);
   pPort->SayPhrase(Arg);
}

void CmdSweep(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   int StartFreq;
   int EndFreq;
   int SweepTime;
   int Sweeps;
   int Level = 0;
   int Converted;

   Converted = sscanf(Arg,"%d %d %d %d %d",
                      &StartFreq,&EndFreq,&SweepTime,&Sweeps,&Level);
   switch(Converted) {
      case 1:
      // Just start frequency, default the rest
         EndFreq = StartFreq;
      case 2:
         SweepTime = 15;
      case 3:
         Sweeps = 1;
      case 4:
         Level = 30000;
         break;
      
      default:
         break;
   }

   if(Level != 0) {
      char Line[80];

      snprintf(Line,sizeof(Line),"TF1 %d EF1 %d TL %d DUR %d RPT %d",
               StartFreq,EndFreq,Level,SweepTime*1000,Sweeps);

      char *pErrPos;
      CToneGen *pGenerator = new CToneGen();

      if(pGenerator->ToneGenSetup(Line,&pErrPos)) {
      // error
         BufPrintf(p,"Error: couldn't parse \"%s\"\r",pErrPos);
         Rcode = TBD_ERR_INVALID_ARG;
      }
      else {
         CPort *pPort = GetPortPtr(CurrentPort);
         pPort->AddToneGen(pGenerator);
      }
   }
   else {
      BufPrintf(p,"Usage: .sweep <start freq> <end freq> <sweep time> <sweeps>"
                  " <sweep level>\r");
   }
}

void CmdDtmfDecode(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CPort *pPort = GetPortPtr(CurrentPort);
   
   pPort->DecodeDTMF(Arg);
}

void CPort::CmdForceID()
{
   SendCWMsg(CWId);
}

void CPort::AddToneGen(CToneGen *pGenerator)
{
   CToneGen *p = NULL;

   if(pGenerator->ID != 0) {
   // Only allow one tone generator of the same id to be running at a time
   // prevents arcade like sounds on weak signals among other annoying things!
      p = ToneGenerators;
      while(p != NULL) {
         if(p->ID == pGenerator->ID) {
         // Duplicate generator
            delete pGenerator;
            break;
         }
         p = p->Link;
      }
   }

   if(p == NULL) {
      pGenerator->Link = ToneGenerators;
      ToneGenerators = pGenerator;
      if(!pAudioC->bWriteWait) {
         WriteAudioOut(pAudioC);
      }
   }
}

void CPort::CancelToneGen(int id)
{
   CToneGen *pToneGen;
   CToneGen *pLastGen;
   CToneGen *pNextGen;

   pToneGen = ToneGenerators;
   pLastGen = NULL;
   while(pToneGen != NULL) {
      pNextGen = pToneGen->Link;
      if(pToneGen->ID == id) {
      // Delete this generator
         if(pLastGen == NULL) {
            ToneGenerators = pNextGen;
         }
         else {
            pLastGen->Link = pNextGen;
         }
         delete pToneGen;
         break;
      }
      else {
         pLastGen = pToneGen;
      }
      pToneGen = pNextGen;
   }
}

void CPort::SetTimeout(struct timeval *pTimer,int Timeout)
{
   long microseconds;

   microseconds = TimeNow.tv_usec + ((Timeout % 1000) * 1000);

   pTimer->tv_usec = microseconds % 1000000;
   pTimer->tv_sec  = TimeNow.tv_sec + (Timeout / 1000) + microseconds / 1000000;
}

void CPort::SendCtone()
{
   if(Ctone != 0 && CtoneMap != 0) do {
   // We have a courtesy tone ID and need to send it to least port
      int Port;
      int Bit = 1;
      CPort *pPort;
      char *ToneSpec = FindToneSpec(Ctone);
      int Bits = CtoneTimer.tv_sec != 0 ? DelayedCtoneMap : CtoneMap;

      DelayedCtoneMap = 0;
      if(ToneSpec == NULL) {
         LOG_ERROR(("%s: Courtesy tone ID #%d is not defined, disabling.\n",
                    __FUNCTION__,Ctone));
         Ctone = 0;
         break;
      }

      for(Port = 0; Port < MAX_PORTS; Port++) {
         if(Bits & 1) {
            pPort = Ports[Port];
            if(pPort != NULL && (pPort->RxMap & MyBit) && !pPort->bTimeOut) {
               if(CtoneDelay != 0 && CtoneTimer.tv_sec == 0) {
               // Set bit for later
                  DelayedCtoneMap |= Bit;
               }
               else {
               // This port is listening to our Rx
                  if(pPort->CtoneResetsTimeout) {
                     TxKeyTime = TimeNow.tv_sec;
                  }

                  CToneGen *pToneGen = new CToneGen;
                  if(pToneGen != NULL) {
                     pToneGen->ToneGenSetup(ToneSpec,NULL);
                     pPort->AddToneGen(pToneGen);
                  }
                  else {
                     LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
                  }
               }
            }
         }
         Bit <<= 1;
         Bits >>= 1;
         if(Bits == 0) {
         // No more ports to test
            break;
         }
      }

      if(DelayedCtoneMap != 0) {
      // Generating a delayed courtesy tone, start the timer
         CtoneTimer = TimeNow;
      }
   } while(FALSE);
}

// return TRUE on error
int CPort::PlayStoredSound(int ID)
{
   int Ret = 0;
   do {
      if(ID == 0) {
         LOG_ERROR(("%s: Called with ID #%d.\n",__FUNCTION__));
         break;
      }
      char *ToneSpec = FindToneSpec(ID);

      if(ToneSpec == NULL) {
         LOG_ERROR(("%s: ID #%d is not defined.\n",__FUNCTION__,ID));
         Ret = TRUE;
         break;
      }
      CToneGen *pGenerator = new CToneGen();

      if(pGenerator != NULL) {
         if(pGenerator->ToneGenSetup(ToneSpec,NULL) == 0) {
            AddToneGen(pGenerator);
         }
      }
      else {
         LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
         Ret = TRUE;
      }
   } while(FALSE);

   return Ret;
}

void CPort::GenVoipCtone()
{
   if(VoipCtone != 0) {
      if(PlayStoredSound(VoipCtone)) {
         LOG_ERROR(("%s: Courtesy tone ID #%d didn't play, disabling.\n",
                    __FUNCTION__,VoipCtone));
         VoipCtone = 0;
      }
   }
}

// Return TRUE if this VoIP client can provide audio
// 
int CPort::VoipAudioAvailable(ConfClient *pCC)
{
   int Ret = FALSE;

   if((pCC->TxMap & MyBit) && !pCC->bRxBuffering) {
   // Talking to us
      if(pCC->pInHead != NULL) {
      // There's an audio buffer available
         Ret = TRUE;
      }
      else if(pCC->pRtpQueue != NULL) {
      // We're not buffering and there's compressed data available for 
      // decompression.  (i.e. we missed a packet and we now need to move on)
         Ret = TRUE;
      }
   }

   return Ret;
}


// Try to find a previously initialized port that's using the same
// serial port as this port
int CPort::FindSerialMaster()
{
   int Ret = FALSE;
#if 0
   int i;
   CPort *pPort;

   for(i = 0; i < PortNum; i++) {
      if((pPort = Ports[i]) != NULL) {
         if(strcmp(pPort->SerialDevName,SerialDevName) == 0 && 
            pPort->SerialSlave == NULL)
         {  // found a port already polling this serial port
            // register us as a slave
            pPort->SerialSlave = this;
            SerialFd = pPort->SerialFd;
            Ret = TRUE;
            break;
         }
      }
   }
#endif
   return Ret;
}

void CmdForceID(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CPort *pPort = GetPortPtr(CurrentPort);

   pPort->CmdForceID();
}

void CmdSayStatus(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   ConfClient *pCC;
   int Connections = piLinkConf->ConfTree->avl_count;
   int Announced = 0;
   struct avl_traverser avl_trans;
   CPort *pPort = GetPortPtr(CurrentPort);
   
   if(Connections > 0) {
      pPort->SayPhrase("I am currently connected to ");
      if(Connections > 1) {
         pPort->SayPhrase("the following stations ");
      }
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(Connections > 1 && (Connections - Announced) == 1) {
            pPort->SayPhrase("and ");
         }
         pPort->SayCall(pCC->Callsign);
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
   }

   pPort->SayPhrase("I have been running for ");
   pPort->SayPhrase(TimeLapse2String(BootTime,FALSE));
}

void ScriptExitCB(int Pid, int Status)
{  
   // Nothing to do, this is just here to prevent creating a zombie army
}

void CmdScript(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
#define MAX_ARGS  16
   char *argv[MAX_ARGS];
   pid_t Pid = -1;
   char  *cp = Arg;
   int   i = 0;

// NB: make sure all buffered streams are written to disk,
// otherwise the data will be written twice, once by the main
// process and once by the forked process !

   if(fflush(NULL)) {
      LOG_ERROR(("%s: fflush failed %s",__FUNCTION__,Err2String(errno)));
   }

   while(i < MAX_ARGS && cp != 0) {
      argv[i++] = cp;
      if((cp = strchr(cp,' ')) != NULL) {
         *cp++ = 0;
         while(isspace(*cp)) {
            cp++;
         }
      }
   }

   if(i == MAX_ARGS) {
      BufPrintf(p,"too many arguments\r");
      LOG_ERROR(("%s: error: too many arguments\n",__FUNCTION__));
   }
   else {
      argv[i] = NULL;
      if((Pid = fork()) < 0) {
         LOG_ERROR(("%s: fork failed %s",__FUNCTION__,Err2String(errno)));
      }
      else if(Pid == 0) {
         int Error;
      // child
         Error = execvp(argv[0],argv);
         exit(1);
      }
      else {
      // Parent
         RegisterExitCB(Pid,ScriptExitCB);
      }
   }
#undef MAX_ARGS
}

// <To> <From> [<From>] ...
// <To> and <From> can be a Voip connection (callsign) or a port name
void DoLinks(ClientInfo *p,char *Arg,int bLink,int bMonitor)
{
#define MAX_ARGS  16
   char *PortNames[MAX_ARGS];
   ConfClient *pVoipClients[MAX_ARGS];
   int TargetPort[MAX_ARGS];
   char  *cp = Arg;
   int i;
   int ArgCount = 0;
   CPort *pFromPort;
   CPort *pToPort;
   ConfClient *pFromCC;
   ConfClient *pToCC;
   CPort *pPort;
   ConfClient *pCC;
   int Port;
   struct avl_traverser avl_trans;

   memset(pVoipClients,0,sizeof(pVoipClients));
   memset(TargetPort,0xff,sizeof(TargetPort));

   do {
      while(ArgCount < MAX_ARGS && *cp) {
         PortNames[ArgCount++] = cp;
         while(*cp && !(isspace(*cp) || *cp == ',')) {
            cp++;
         }

         while(*cp && (isspace(*cp) || *cp == ',')) {
            *cp++ = 0;
         }
      }

      if(ArgCount == MAX_ARGS) {
         BufPrintf(p,"Error: too many arguments\r",cp);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      for(i = 0; i < ArgCount; i++) {
      // Find specified port or Voip connection
         for(Port = 0; Port < NumPorts; Port++) {
            if((pPort = Ports[Port]) != NULL && 
               STRCMP(pPort->PortName,PortNames[i]) == 0) 
            {  // found port
               TargetPort[i] = Port;
               break;
            }
         }

         if(Port == NumPorts) {
         // Not an Port, look for an Voip connection
            pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
            while(pCC != NULL) {
               if(STRCMP(pCC->Callsign,PortNames[i]) == 0) {
               // Found VoIP connection
                  pVoipClients[i] = pCC;
                  break;
               }
               pCC = (ConfClient *) avl_t_next(&avl_trans);
            }
         }

         if(TargetPort[i] == -1 && pVoipClients[i] == NULL) {
         // port and connection not found
            BufPrintf(p,"Error: can't find port or connection named \"%s\"\r",
                      PortNames[i]);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }
      }

      if(Rcode != 0) {
         break;
      }

      pToCC = pVoipClients[0];
      pToPort = TargetPort[0] == -1 ? NULL : GetPortPtr(TargetPort[0]);

      if(pToCC == NULL && pToPort == NULL) {
         LOG_ERROR(("%s#%d: Internal error!\n",__FUNCTION__,__LINE__));
         BufPrintf(p,"Internal error @ %s#%d\r",__FUNCTION__,__LINE__);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      if(ArgCount == 1 && !bLink) {
      // unlink everything from specified port/connection
         if(pToPort != NULL) {
         // Argument is a port, unlink everything from it
            pToPort->RxMap = 0;
            for(Port = 0; Port < NumPorts; Port++) {
               if(Ports[Port] != NULL) {
                  Ports[Port]->RxMap &= ~pToPort->MyBit;
               }
            }

            pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
            while(pCC != NULL) {
               pCC->TxMap &= ~pToPort->MyBit;
               pCC->RxMap &= ~pToPort->MyBit;
               pCC = (ConfClient *) avl_t_next(&avl_trans);
            }
            BufPrintf(p,"all ports and connections unlinked from %s\r",
                      PortNames[0]);
         }
         else {
         // Argument is a voip connection
            pToCC->RxMap = pToCC->TxMap = 0;
            BufPrintf(p,"all ports unlinked from %s\r",PortNames[0]);
         }
         break;
      }

      if(ArgCount < 2) {
         BufPrintf(p,"Error: not enough arguments\r");
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      BufPrintf(p,"%slinking ",bLink ? "" : "un");

      for(i = 1; i < ArgCount; i++) {
         if(TargetPort[i] != -1) {
         // "From" argument is a port
            pFromPort = GetPortPtr(TargetPort[i]);
            BufPrintf(p,"%s%s",i > 1 ? ", " : "",pFromPort->PortName);
            if(pToPort != NULL) {
            // "To" argument is a port
               if(bLink) {
               // link From -> To
                  pToPort->RxMap |= pFromPort->MyBit;
                  if(!bMonitor) {
                  // not monitor mode, also link To->From
                     pFromPort->RxMap |= pToPort->MyBit;
                  }
                  else {
                  // monitor mode, also unlink link To->From
                     pFromPort->RxMap &= ~pToPort->MyBit;
                  }
               }
               else {
               // Unlink From & To
                  pToPort->RxMap &= ~pFromPort->MyBit;
                  pFromPort->RxMap &= ~pToPort->MyBit;
               }
            }
            else {
            // "To" argument is a Voip Connection
               if(bLink) {
               // link From -> To
                  pToCC->RxMap |= pFromPort->MyBit;
                  if(!bMonitor) {
                  // not monitor mode, also link To->From
                     pToCC->TxMap |= pFromPort->MyBit;
                  }
                  else {
                  // monitor mode, also unlink To->From
                     pToCC->TxMap &= ~pFromPort->MyBit;
                  }
               }
               else {
               // Unlink From & To
                  pToCC->RxMap &= ~pFromPort->MyBit;
                  pToCC->TxMap &= ~pFromPort->MyBit;
               }
            }
         }
         else if(pVoipClients[i] != NULL) {
         // "From" argument is a VoIP client
            pFromCC = pVoipClients[i];
            BufPrintf(p,"%s%s",i > 1 ? ", " : "",pFromCC->Callsign);
            if(pToPort != NULL) {
            // "To" argument is a port
               if(bLink) {
               // Link From -> To
                  pFromCC->TxMap |= pToPort->MyBit;
                  if(!bMonitor) {
                     pFromCC->RxMap |= pToPort->MyBit;
                  }
                  else {
                     pFromCC->RxMap &= ~pToPort->MyBit;
                  }
               }
               else {
               // Unlink From & To
                  pFromCC->TxMap &= ~pToPort->MyBit;
                  pFromCC->RxMap &= ~pToPort->MyBit;
               }
            }
            else {
            // "To" argument is a Voip Connection
               p->Count = 0;  // reinit for BufPrintf to toss previous output
               BufPrintf(p,"%c" NDATA,ILINK_DATA_PACKET);
               BufPrintf(p,"Error: VoIP connnections can not be unlinked "
                         "from each other");
               Rcode = TBD_ERR_INVALID_ARG;
               break;
            }
         }
      }
      if(bLink) {
         BufPrintf(p," to %s%s\r",PortNames[0],bMonitor ? " (monitor only)":"");
      }
      else {
         BufPrintf(p," from %s\r",PortNames[0]);
      }
   } while(FALSE);
#undef MAX_ARGS
}

void CmdLink(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   int RxMap;
   int i;
   int j;
   int bFirst;
   int bFirstVoIP;
   int Bits;
   CPort *pFrom;
   CPort *pTo;
   ConfClient *pCC;
   struct avl_traverser avl_trans;

   if(*Arg) {
      if(Arg[0] == '-') {
         switch(Arg[1]) {
            case 'm':
            // Monitor command
               Arg += 2;
               while(*Arg == ' ') {
                  Arg++;
               }
               DoLinks(p,Arg,TRUE,TRUE);
               break;

            case 'h':
            case '?':
               BufPrintf(p,"Usage: .link [-m] <destination> <source> "
                            "[<source>...]\r");
               Rcode = TBD_ERR_INVALID_ARG;
               break;

            default:
               BufPrintf(p,"Error: Unknown switch \"%c\"\r",Arg[1]);
               Rcode = TBD_ERR_INVALID_ARG;
               break;
         }
      }
      else {
         DoLinks(p,Arg,TRUE,FALSE);
      }
   }
   else {
   // No argument, list linked ports
      BufPrintf(p,"Port mapping:\n");
      for(i = 0; i < NumPorts; i++) {
         if((pTo = Ports[i]) != NULL) {
            BufPrintf(p," %s: ",pTo->PortName);
            bFirst = TRUE;
            if((RxMap = Ports[i]->RxMap) != 0) {
               for(j = 0; j < NumPorts; j++) {
                  if((pFrom = Ports[j]) != NULL) {
                     if(RxMap & 1) {
                        BufPrintf(p,"%s%s",bFirst ? "" : ", ",pFrom->PortName);
                        bFirst = FALSE;
                     }
                     RxMap >>= 1;
                     if(RxMap == 0) {
                        break;
                     }
                  }
               }
            }

         // List VoIP connections for this port
            pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
            while(pCC != NULL) {
               if((pCC->TxMap & pTo->MyBit)  && !pCC->bKicked) {
                  BufPrintf(p,"%s%s",bFirst ? "" : ", ",pCC->Callsign);
                  bFirst = FALSE;
               }
               pCC = (ConfClient *) avl_t_next(&avl_trans);
            }
            BufPrintf(p,"%s\r",bFirst ? "none" : "");
         }
      }
   // List VoIP connections
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);

      bFirstVoIP = TRUE;
      while(pCC != NULL) {
         if(!pCC->bKicked) {
            if(bFirstVoIP) {
               bFirstVoIP = FALSE;
               BufPrintf(p,"Voip mappings:\n");
            }
            BufPrintf(p," %s: ",pCC->Callsign);
            bFirst = TRUE;
            if((Bits = pCC->RxMap) != 0) {
               int Port;
               for(Port = 0; Port < NumPorts; Port++) {
                  if(Bits & 1) {
                     BufPrintf(p,"%s%s",bFirst ? "" : ", ",
                               GetPortPtr(Port)->PortName);
                     bFirst = FALSE;
                  }
                  Bits >>= 1;
                  if(Bits == 0) {
                  // No more ports to test
                     break;
                  }
               }
               BufPrintf(p,"\r");
            }
            else {
               BufPrintf(p,"none\r");
            }
         }
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
   }
}

void CmdUnLink(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   int i;
   ConfClient *pCC;
   struct avl_traverser avl_trans;

   if(*Arg == 0) {
   // No argument
      BufPrintf(p,"usage:\r  .unlink all\r  .unlink rf\r  .unlink voip\r");
      BufPrintf(p,"  .unlink <destination> [<source>...]\r");
      Rcode = TBD_ERR_INVALID_ARG;
   }
   else if(STRCMP(Arg,"rf") == 0) {
      for(i = 0; i < NumPorts; i++) {
         if(Ports[i] != NULL) {
            Ports[i]->RxMap = 0;
         }
      }
      BufPrintf(p,"all ports unlinked\r");
   }
   else if(STRCMP(Arg,"voip") == 0) {
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         pCC->TxMap = pCC->RxMap = 0;
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
      BufPrintf(p,"all voip connections unlinked\r");
   }
   else if(STRCMP(Arg,"all") == 0) {
      for(i = 0; i < NumPorts; i++) {
         if(Ports[i] != NULL) {
            Ports[i]->RxMap = 0;
         }
      }
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         pCC->TxMap = pCC->RxMap = 0;
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
      BufPrintf(p,"all voip connections and ports unlinked\r");
   }
   else {
      DoLinks(p,Arg,FALSE,FALSE);
   }
}

char *FindToneSpec(int ID)
{
   char *Ret = NULL;
   SavedToneSpec *pSpec = ToneSpecs;

   while(pSpec != NULL) {
      if(pSpec->ID == ID) {
      // Match
         Ret = pSpec->ToneSpec;
         break;
      }
      pSpec = pSpec->Link;
   }
   return Ret;
}

